###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################

from flask import Flask as Namespace, request, abort, make_response, Response, send_from_directory, safe_join, jsonify, render_template_string, send_file  # noqa


# based on https://github.com/miguelgrinberg/oreilly-flask-apis-video/blob/master/orders/app/decorators/caching.py
def cache_control(*directives):
  """Insert a Cache-Control header with the given directives."""
  import functools

  def decorator(f):
    @functools.wraps(f)
    def wrapped(*args, **kwargs):
      # invoke the wrapped function
      rv = f(*args, **kwargs)

      # convert the returned value to a response object
      rv = make_response(rv)

      # insert the Cache-Control header and return response
      rv.headers['Cache-Control'] = ', '.join(directives)
      return rv

    return wrapped

  return decorator


def no_cache(f):
  """Insert a no-cache directive in the response. This decorator just
  invokes the cache-control decorator with the specific directives."""
  return cache_control('private', 'no-cache', 'no-store', 'max-age=0')(f)


def etag(f):
  """Add entity tag (etag) handling to the decorated route."""
  import functools

  @functools.wraps(f)
  def wrapped(*args, **kwargs):
    if request.method not in ['GET', 'HEAD']:
      # etags only make sense for request that are cacheable, so only
      # GET and HEAD requests are allowed
      return f(*args, **kwargs)

    # invoke the wrapped function and generate a response object from
    # its result
    rv = f(*args, **kwargs)
    rv = make_response(rv)

    # if the response is not a code 200 OK then we let it through
    # unchanged
    if rv.status_code != 200 or rv.direct_passthrough or not rv.implicit_sequence_conversion:
      return rv

    rv.add_etag()
    return rv.make_conditional(request)
  return wrapped



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str
from builtins import range as number_range
from builtins import object
import itertools
from functools import reduce, cmp_to_key
from numpy import NaN, isnan

all_f = all


def fix(v, size=0):
  return v if v >= 0 else (size + 1 + v)


class SingleRangeElem(object):
  def __init__(self, val):
    self.start = val

  @property
  def end(self):
    return self.start + 1

  @property
  def step(self):
    return 1

  def asslice(self):
    return self.start

  @property
  def isall(self):
    return False

  @property
  def issingle(self):
    return True

  @property
  def isunbound(self):
    return False

  def __len__(self):
    return 1

  @staticmethod
  def size(size=0):
    return 1

  def reverse(self):
    return SingleRangeElem(self.start)

  def invert(self, index, size=0):
    return fix(self.start, size) + index

  def __iter__(self):
    return self.iter()

  def iter(self, size=0):
    return iter([self.start])

  def contains(self, value, size=0):
    return fix(self.start, size) == value

  def __in__(self, value):
    return self.contains(value)

  def __str__(self):
    return str(self.start)

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.start == other.start
    return False

  def copy(self):
    return self.__copy__()

  def __copy__(self):
    return SingleRangeElem(self.start)


class RangeElem(object):
  def __init__(self, start, end=-1, step=1):
    self.start = start
    self.end = end
    self.step = step

  @property
  def isall(self):
    return self.start == 0 and self.end == -1 and self.step == 1

  def asslice(self):
    return slice(self.start, self.end, self.step)

  @property
  def issingle(self):
    return (self.start + self.step) == self.end

  @property
  def isunbound(self):
    return self.start < 0 or self.end < 0

  @staticmethod
  def all():
    return RangeElem(0)

  @staticmethod
  def none():
    return RangeElem(0, 0)

  @staticmethod
  def single(val):
    return SingleRangeElem(val)

  @staticmethod
  def range(start, end=-1, step=1):
    if (start + step) == end:
      return RangeElem.single(start)
    return RangeElem(start, end, step)

  def __len__(self):
    return self.size()

  def size(self, size=NaN):
    t = fix(self.end, size)
    f = fix(self.start, size)
    if self.step == 1:
      return max(t - f, 0)
    elif self.step == -1:
      if self.end == -1:
        return max(f - -1, 0)
      return max(f - t, 0)
    d = t - f + 1 if self.step > 0 else f - t + 1
    s = abs(self.step)
    if d <= 0:
      return 0
    return (d // s)

  def reverse(self):
    if self.start > 0:
      t = self.start - 1
      f = self.end - 1
      return RangeElem(f, t, - self.step)
    else:  # step < 0
      t = self.start - 1
      f = self.end - 1
      return RangeElem(f, t, - self.step)

  def invert(self, index, size=0):
    if self.isall:
      return index
    return fix(self.start, size) + index * self.step

  def __iter__(self):
    return self.iter()

  def iter(self, size=0):
    if self.step < 0 and self.end == -1:
      # keep negative to have 0 included
      return iter(number_range(fix(self.start, size), -1, self.step))
    return iter(number_range(fix(self.start, size), fix(self.end, size), self.step))

  def contains(self, value, size=NaN):
    if self.isall:
      return True
    f = fix(self.start, size)
    t = fix(self.end, size)
    if self.step == -1:
      if self.end == -1:
        return 0 <= value <= f
      return (value <= f) and (value > t)
    elif self.step == 1:
      return (value >= f) and (value < t)
    else:
      return value in list(self.iter(size))

  def __in__(self, value):
    return self.contains(value)

  def __str__(self):
    if self.isall:
      return ''
    if self.issingle:
      return str(self.start)
    r = str(self.start) + ':' + str(self.end)
    if self.step != 1:
      r = r + ':' + str(self.step)
    return r

  def __eq__(self, other):
    if isinstance(other, self.__class__):
      return self.start == other.start and self.end == other.end and self.step == other.step
    return False

  def copy(self):
    return self.__copy__()

  def __copy__(self):
    return RangeElem(self.start, self.end, self.step)

  @staticmethod
  def parse(code):
    if len(code) == 0:
      return RangeElem.all()

    def parse_elem(v, default_value=None):
      v = v.strip()
      if len(v) == 0 and default_value is not None:
        return default_value
      try:
        return int(v)
      except ValueError:
        raise Exception('parse error: "' + v + '" is not a valid integer')

    parts = code.split(':')
    if len(parts) == 1:
      return RangeElem.single(parse_elem(parts[0]))
    elif len(parts) == 2:
      return RangeElem(parse_elem(parts[0], 0), parse_elem(parts[1], -1))
    elif len(parts) == 3:
      return RangeElem(parse_elem(parts[0], 0), parse_elem(parts[1], -1), parse_elem(parts[2], 1))
    raise Exception('parse error: "' + code + '" is not a valid range specifier')


class Range1D(object):
  def __init__(self, arg=None):
    if isinstance(arg, list):
      self._elems = arg
    elif isinstance(arg, Range1D):
      self._elems = arg._elems
    else:
      self._elems = []

  def __len__(self):
    return self.size()

  def copy(self):
    return Range1D(self._elems[:])

  def __copy__(self):
    return self.copy()

  def size(self, size=NaN):
    if isnan(size) and self.isunbound:
      return NaN
    return reduce(lambda s, x: s + x.size(size), self._elems, 0)

  @staticmethod
  def all():
    return Range1D([RangeElem.all()])

  @staticmethod
  def single(item):
    return Range1D([RangeElem.single(item)])

  @staticmethod
  def none():
    return Range1D()

  @staticmethod
  def from_list(indices):
    return Range1D(Range1D._compress(indices))

  @staticmethod
  def _compress(indices):
    l = len(indices)
    if l == 0:
      return []
    elif l == 1:
      return [RangeElem.single(indices[0])]
    r = []
    deltas = [e - indices[i] for i, e in enumerate(indices[1:])]
    start = 0
    act = 1

    while act < l:
      while act < l and deltas[start] == deltas[act - 1]:  # while the same delta
        act += 1
      if act == start + 1:  # just a single item used
        r.append(RangeElem.single(indices[start]))
      else:
        # +1 since end is excluded
        # fix while just +1 is allowed and -1 is not allowed
        if deltas[start] == 1:
          r.append(RangeElem.range(indices[start], indices[act - 1] + deltas[start], deltas[start]))
        else:
          for i in number_range(start, act):
            r.append(RangeElem.single(indices[i]))
      start = act
      act += 1
    while start < len(indices):  # corner case by adding act+1, it might happened that last one isn't considered
      r.append(RangeElem.single(indices[start]))
      start += 1
    return r

  @property
  def isall(self):
    return len(self._elems) == 1 and self[0].isall

  @property
  def isnone(self):
    return len(self._elems) == 0

  @property
  def isunbound(self):
    return any((d.isunbound for d in self._elems))

  @property
  def _islist(self):
    return not any(not d.issingle for d in self._elems)

  def append(self, *args):
    def convert(p):
      if isinstance(p, str):
        return RangeElem.parse(p)
      elif isinstance(p, int):
        return RangeElem.single(p)
      elif isinstance(p, list):
        return RangeElem.range(p[0], p[1], p[2])
      return p

    self._elems.extend((convert(p) for p in args))

  def push_slice(self, start, end=-1, step=1):
    self._elems.append(RangeElem(start, end, step))

  def push_list(self, indices):
    self._elems.extend(Range1D._compress(indices))

  def set_slice(self, start, end=-1, step=1):
    self._elems = []
    self.push_slice(start, end, step)

  def set_list(self, indices):
    self._elems = []
    self.push_list(indices)

  def __getitem__(self, i):
    if i < 0:
      i += len(self._elems)
    if i < 0 or i >= len(self._elems):
      return RangeElem.none()
    return self._elems[i]

  def asslice(self, no_ellipsis=False):
    if self.isall:
      return slice(0, -1) if no_ellipsis else Ellipsis
    if self.isnone:
      return []
    if len(self._elems) == 1:
      return self._elems[0].asslice()
    return self.tolist()

  @property
  def is_identity_range(self):
    return len(self._elems) == 1 and self._elems[0].start == 0 and self._elems[0].step == 1

  def repeat(self, ntimes=1):
    if ntimes == 1:
      return self
    r = []
    for i in number_range(ntimes):
      r.extend(self._elems)
    return Range1D(r)

  def pre_multiply(self, sub, size=0):
    if self.isall:
      return sub
    if sub.isall:
      return self
    if self.is_identity_range:  # identity lookup
      return sub

    # TODO optimize
    l = list(self.iter(size))
    s = sub.iter(len(l))
    r = []
    while s.hasNext():
      i = next(s)
      if 0 <= i < len(l):  # check for out of range
        r.append(l[i])

    return Range1D.from_list(r)

  def union(self, other, size=0):
    if self.isall or other.isnone:
      return self

    if other.isall or self.isnone:
      return other

    r = list(self.iter(size))
    it2 = other.iter(size)
    for i in it2:
      if i not in r:
        r.append(i)

    return Range1D.from_list(sorted(r))

  def intersect(self, other, size=0):
    if self.isnone or other.isnone:
      return Range1D.none()

    if self.isall:
      return other

    if other.isall:
      return self

    it1 = list(self.iter(size))
    it2 = other.iter(size)
    r = [i for i in it2 if i in it1]
    return Range1D.from_list(sorted(r))

  def without(self, without, size=0):
    if self.isnone or without.isnone:
      return self.copy()

    if without.isall:
      return Range1D.none()

    it1 = self.iter(size)
    it2 = list(without.iter(size))
    r = [i for i in it1 if i not in it2]
    return Range1D.from_list(sorted(r))

  def invert(self, index, size=0):
    if self.isall:
      return index

    if self.isnone:
      return -1  # not mapped

    act = 0
    s = self._elems[0].size(size)
    total = s
    while total > index and act < len(self):
      act += 1
      s = self._elems[act].size(size)
      total += s

    if act >= len(self._elems):
      return -1  # not mapped

    return self._elems[act - 1].invert(index - total + s, size)

  def index(self, *args):
    if len(args) == 0:
      return []

    if isinstance(args[0], Range1D):
      return self.index_range_of(args[0], args[1])
    base = list(self.iter())
    if len(args) == 1:
      if type(args[0]) is int:
        return base.index(args[0])
      arr = args[0]
    else:
      arr = args
    return [base.index(index) for index in arr]

  def index_range_of(self, r, size=0):
    if r.isnone or self.isnone:
      return Range1D.none()
    if self.is_identity_range:
      end = self._elems[0].end
      result = [d for d in r if 0 <= d < end]
    else:
      arr = list(self.iter())
      result = [arr.index(d) for d in arr if d in arr]
    return Range1D.from_list(result)

  def filter(self, data, size, transform=lambda x: x):
    if self.isall:
      return [transform(x) for x in data]
    return [transform(data[i]) for i in self.iter(size)]

  def iter(self, size=0):
    if self._islist:
      return (d.start for d in self._elems)
    else:
      return itertools.chain(*[d.iter(size) for d in self._elems])

  def __iter__(self):
    return self.iter()

  def tolist(self, size=0):
    return list(self.iter(size))

  def contains(self, value, size=0):
    return any(elem.contains(value, size) for elem in self._elems)

  def sort(self, cmp):
    arr = list(self.iter())
    arr.sort(key=cmp_to_key(cmp))
    return Range1D.from_list(arr)

  def remove_duplicates(self, size=0):
    arr = list(self.iter())
    arr.sort()
    arr = [di for i, di in enumerate(arr) if di != arr[i - 1]]  # same value as before, remove
    return Range1D.from_list(arr)

  def reverse(self):
    a = [r.reverse() for r in self._elems]
    a = a.reverse()
    return Range1D(a)

  def __str__(self):
    if self.isall:
      return ''

    if len(self) == 1:
      return str(self._elems[0])

    return '(' + ','.join(str(e) for e in self._elems) + ')'


class Range1DGroup(Range1D):
  def __init__(self, name, color, base=None):
    super(Range1DGroup, self).__init__(base)
    self.name = name
    self.color = color

  def pre_multiply(self, sub, size=0):
    r = super(Range1DGroup, self).pre_multiply(sub, size)
    return Range1DGroup(self.name, self.color, r)

  def union(self, other, size=0):
    r = super(Range1DGroup, self).union(other, size)
    return Range1DGroup(self.name, self.color, r)

  def intersect(self, other, size=0):
    r = super(Range1DGroup, self).intersect(other, size)
    return Range1DGroup(self.name, self.color, r)

  def without(self, without, size=0):
    r = super(Range1DGroup, self).without(without, size)
    return Range1DGroup(self.name, self.color, r)

  def sort(self, cmp):
    r = super(Range1DGroup, self).sort(cmp)
    return Range1DGroup(self.name, self.color, r)

  def __str__(self):
    return '"' + self.name + '""' + self.color + '"' + str(super(Range1DGroup, self))


def as_ungrouped(range):
  return Range1DGroup('unnamed', 'gray', range)


def composite(name, groups):
  return CompositeRange1D(name, groups)


def to_base(groups):
  if len(groups) == 1:
    return groups[0]
  r = groups[0].tolist()
  for g in groups[1:]:
    r.extend((i for i in g if i not in r))
  return Range1D.from_list(r)


class CompositeRange1D(Range1D):
  def __init__(self, name, groups, base=None):
    super(CompositeRange1D, self).__init__(base if base is not None else to_base(groups))
    self.name = name
    self.groups = groups

  def pre_multiply(self, sub, size=0):
    r = super(CompositeRange1D, self).pre_multiply(sub, size) if len(self.groups) > 1 else None
    return CompositeRange1D(self.name, [g.pre_multiply(sub, size) for g in self.groups], r)

  def union(self, other, size=0):
    r = super(CompositeRange1D, self).union(other, size) if len(self.groups) > 1 else None
    return CompositeRange1D(self.name, [g.union(other, size) for g in self.groups], r)

  def intersect(self, other, size=0):
    r = super(CompositeRange1D, self).intersect(other, size) if len(self.groups) > 1 else None
    return CompositeRange1D(self.name, [g.intersect(other, size) for g in self.groups], r)

  def without(self, without, size=0):
    r = super(CompositeRange1D, self).without(without, size) if len(self.groups) > 1 else None
    return CompositeRange1D(self.name, [g.without(without, size) for g in self.groups], r)

  def sort(self, cmp):
    r = super(CompositeRange1D, self).sort(cmp) if len(self.groups) > 1 else None
    return CompositeRange1D(self.name, [g.sort(cmp) for g in self.groups], r)

  def __str__(self):
    return '"' + self.name + '":' + ','.join((str(g) for g in self.groups)) + ''


class Range(object):
  def __init__(self, dims=[]):
    self.dims = dims

  @property
  def isall(self):
    return all_f((d.isall for d in self.dims))

  @property
  def isnone(self):
    return all_f((d.isnone for d in self.dims))

  @property
  def ndim(self):
    return len(self.dims)

  def __getitem__(self, item):
    if len(self.dims) > item:
      return self.dims[item]
    for i in number_range(len(self.dims), item + 1):
      self.dims.append(Range1D.all())
    return self.dims[item]

  def __setitem__(self, key, value):
    self.dims[key] = value

  def __len__(self):
    return len(self.dims)

  def __eq__(self, other):
    if self is other or (self.isall and other.isall) or (self.isnone or other.isnone):
      return True
    return str(self) == str(other)

  def pre_multiply(self, other, size=[]):
    if self.isall:
      return other.copy()
    if other.isall:
      return self.copy()
    return Range([d.pre_multiply(other[i], size[i] if i >= len(size) else 0) for i, d in enumerate(self.dims)])

  def union(self, other, size=[]):
    if self.isall or other.isnone:
      return self.copy()
    if other.isall or self.isnone:
      return other.copy()
    return Range([d.union(other[i], size[i] if i >= len(size) else 0) for i, d in enumerate(self.dims)])

  def intersect(self, other, size=[]):
    if self.isnone or other.isnone:
      return none()
    if self.isall:
      return other.copy()
    if other.isall:
      return self.copy()

    return Range([d.intersect(other[i], size[i] if i >= len(size) else 0) for i, d in enumerate(self.dims)])

  def without(self, without, size=[]):
    if self.isnone or without.isnone:
      return self.copy()
    if without.isall:
      return none()

    return Range([d.without(without[i], size[i] if i >= len(size) else 0) for i, d in enumerate(self.dims)])

  def copy(self):
    return Range([d.copy() for d in self.dims])

  def asslice(self, no_ellipsis=False):
    if self.isall:
      return Ellipsis
    return tuple((d.asslice(no_ellipsis) for d in self.dims))

  def swap(self):
    a = [d.copy() for d in self.dims]
    a.reverse()
    return Range(a)

  def filter(self, data, size=[]):
    if self.isall:
      return data

    ndim = self.ndim

    # recursive variant for just filtering the needed rows
    def filter_dim(i):
      if i >= ndim:
        return lambda x: x
      d = self[i]
      nex = filter_dim(i + 1)
      s = size[i] if len(size) > i else 0
      return lambda elem: d.filter(elem, s, nex) if isinstance(elem, list) else elem

    f = filter_dim(0)
    return f(data)

  def invert(self, indices, size=[]):
    if self.isall:
      return indices
    return [self.dim(i).invert(index, size[i] if len(size) > i else 0) for i, index in enumerate(indices)]

  def index_of_range(self, r, size=[]):
    if r.isnone or self.isnone:
      return none()
    if self.isnone or r.isall:
      return self.copy()
    return Range([d.index_range_of(r[i], size[i] if len(size) > i else 0) for i, d in enumerate(self.dims)])

  def index(self, index_or_range, *args):
    if type(index_or_range) is Range:
      return self.index_of_range(index_or_range, args[0])
    if len(args) == 0:
      if type(index_or_range) is int:
        return self[0].index(index_or_range)
      arr = index_or_range
    else:
      arr = [index_or_range]
      arr.extend(args)
    if len(arr) == 0:
      return []
    return [self[i].index(index) for i, index in enumerate(arr)]

  def size(self, size=[]):
    if self.isall:
      return size
    return [r.size(size[i] if len(size) > i else 0) for i, r in enumerate(self.dims)]

  def split(self):
    return [Range([dim]) for dim in self.dims]

  def __str__(self):
    return ','.join((str(d) for d in self.dims))

  def __iter__(self):
    return iter(self.dims)


def all():
  return Range()


def none():
  return Range([Range1D.none(), Range1D.none()])


def from_slice(start, end=-1, step=1):
  r = Range()
  r[0].set_slice(start, end, step)
  return r


def range(*args):
  if len(args) == 0:
    return all()
  r = Range()
  if isinstance(args[0], list):
    for i, arr in enumerate(args):
      if len(arr) == 0:
        continue
      r[i].set_slice(arr[0], arr[1], arr[2])
  if type(args[0]) is int:
    r[0].set_slice(args[0], args[1], args[2])

  return r


def join(*args):
  if len(args) == 0:
    return all()
  r = Range()
  if isinstance(args[0], list):
    args = args[0]
  r.dims = [ri[0] for ri in args]
  return r


def from_list(*args):
  if len(args) == 0:
    return all()
  r = Range()
  if isinstance(args[0], list) and type(args[0][0]) is Range1D:
    r.dims = args[0]
  elif isinstance(args[0], list):  # array mode
    for i, arr in enumerate(args):
      if type(arr) is Range1D:
        r[i] = arr
      else:
        r[i].set_list(arr)
  elif type(args[0]) is int:  # single slice mode
    r[0].set_list(args)
  elif type(args[0]) is Range1D:
    r.dims = args
  return r


# Range EBNF grammar
# R   = Dim : ',' Dim
# Dim = '' | SR | '(' SR : ',' SR '  ')'
# SR  = N [ ':' N [ ':' N ] ]
# N   = '0'...'9'
# Str =  '"' literal '"'
# Name= Str
# Col = Str
# GDim= Name Col Dim
# CDim= Name ':' GDim : ',' GDim  ''
def parse_range(code):
  act = 0
  dims = []
  code = code.strip()
  while act < len(code):
    c = code[act]
    if c == '"':
      act, dim = parse_named_range1d(code, act)
      act += 1  # skip ,
      dims.append(dim)
    elif c == ',':
      act += 1
      dims.append(Range1D.all())
    else:
      ract, dim = parse_range1d(code, act)
      act = ract + 1  # skip ,
      dims.append(dim)
  return Range(dims)


def parse_named_range1d(code, act):
  act += 1  # skip "
  end = code.index('"', act)
  name = code[act:end]
  act = end + 1
  c = code[act]
  if c == '"':
    end = code.index('"', act + 1)
    ract, dim = parse_range1d(code, end + 1)
    return ract, Range1DGroup(name, code[act + 1:end], dim)
  elif c == '{':
    groups = []
    while code[act] != '}':
      ract, dim = parse_named_range1d(code, act + 1)
      groups.append(dim)
      act = ract
    return act + 1, CompositeRange1D(name, groups)
  else:  # error
    return act, Range1D.all()


def parse_range1d(code, act):
  if act >= len(code):
    return act, Range1D.all()
  c = code[act]
  if c == ',' or c == '{':
    n = act
    r = Range1D.all()
  elif c == '(':
    n = code.index(')', act)
    r = Range1D([RangeElem.parse(ni) for ni in code[act + 1: n].split(',')])
    n += 1
  else:
    n = code.find(',', act)
    n2 = code.find('}', act)
    if n >= 0 and n2 >= 0:
      n = min(n, n2)
    elif n < 0:
      n = n2

    if n < 0:
      n = len(code)
    r = Range1D([RangeElem.parse(code[act:n])])
  return n, r


def parse(*args):
  if len(args) == 0:
    return all()
  if len(args) == 1 and type(args[0]) is Range:
    return args[0]
  return parse_range(','.join(args))



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import next
from . import ns
import os
import re

import logging

_log = logging.getLogger(__name__)


black_list = re.compile(r'(.*\.(py|pyc|gitignore|gitattributes)|(\w+)/((config|package)\.json|_deploy/.*))')
public_dir = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', 'public'))


def _is_on_black_list(path):
  # print 'check',path,black_list.match(path) is not None
  return black_list.match(path) is not None


def _deliver_production(path):
  # print path
  if path.endswith('/'):
    path += 'index.html'
  if _is_on_black_list(path):
    return 'This page does not exist', 404
  # serve public
  return ns.send_from_directory(public_dir, path)


def _deliver(path):
  # print path
  if path.endswith('/'):
    path += 'index.html'
  if _is_on_black_list(path):
    return 'This page does not exist', 404

  # serve public
  if os.path.exists(ns.safe_join(public_dir, path)):
    return ns.send_from_directory(public_dir, path)

  # check all plugins
  elems = path.split('/')
  if len(elems) > 0:
    plugin_id = elems[0]
    elems[0] = 'build'
    from .plugin import plugins
    plugin = next((p for p in plugins() if p.id == plugin_id), None)
    if plugin:
      dpath = ns.safe_join(plugin.folder, '/'.join(elems))
      if os.path.exists(dpath):
        # send_static_file will guess the correct MIME type
        # print 'sending',dpath
        return ns.send_from_directory(plugin.folder, '/'.join(elems))

  return 'This page does not exist', 404


def _generate_index():
  text = ["""
    <!DOCTYPE html><html><head lang="en">
    <meta charset="UTF-8"> <title>Caleydo Web Apps</title>
    <link href="//fonts.googleapis.com/css?family=Yantramanav:700,400" rel="stylesheet" type="text/css">
    <link href="assets/main.css" rel="stylesheet" type="text/css"></head>
    <body><div class="container"> <header>
    <h1><img src="assets/caleydo_text_right.svg" alt="Caleydo" width="200" height="40"> Web Apps</h1> </header>
    <main> <nav id="apps"> <input type="search" id="search" class="search" placeholder="Search App" autocomplete="off"/>
    <div class="keyboard-navigation-hint"> <span>Jump to an app:</span> <span><b>&#8593;</b><b>&#8595;</b> to navigate</span> <span><b>&#8629;</b> to select</span> </div>
    <ul class="list">
    """]

  # filter list and get title for apps
  from .plugin import plugins
  apps = sorted((p for p in plugins() if p.is_app()), key=lambda p: p.title)

  for app in apps:
    text.append('<li>')
    text.append('<a class="appinfo" href="/' + app.id + '/"><span class="title">' + app.title + '</span><span class="name">' + app.name + '</span><span class="description">' + app.description + '</span></a>')
    text.append('<div class="links">')
    if app.homepage and app.homepage != '':
      text.append('<a href="' + app.homepage + '" target="_blank" class="homepage"><span>Visit homepage</span></a>')

    if app.repository and app.repository != '':
      text.append('<a href="' + app.repository + '" target="_blank" class="github"><span>Open repository</span></a>')

    text.append('</div>')
    text.append('</li>')

  text.append("""</ul> </nav> </main> <footer>
      <img src="assets/caleydo_c.svg" alt="Caleydo" width="20" height="20">
      <a href="http://caleydo.org">caleydo.org</a> </footer></div>
      <script src="assets/list.min.js"></script><script src="assets/main.js"></script>
      </body></html>
    """)
  return '\n'.join(text)


def _build_info():
  from codecs import open
  from .plugin import metadata

  dependencies = []
  plugins = []
  build_info = dict(plugins=plugins, dependencies=dependencies)

  requirements = 'requirements.txt'
  if os.path.exists(requirements):
    with open(requirements, 'r', encoding='utf-8') as f:
      dependencies.extend([l.strip() for l in f.readlines()])

  for p in metadata().plugins:
    if p.id == 'phovea_server':
      build_info['name'] = p.name
      build_info['version'] = p.version
      build_info['resolved'] = p.resolved
    else:
      desc = dict(name=p.name, version=p.version, resolved=p.resolved)
      plugins.append(desc)

  return ns.jsonify(build_info)


# health check for docker-compose, kubernetes
def _health():
    return 'ok', 200


def default_app():
  # from .config import view, _initialize
  from phovea_server import config
  # check initialization
  if config._c is None:
    config._initialize()
  app = ns.Namespace(__name__)
  cc = config.view('phovea_server')
  if cc.env.startswith('dev'):
    app.add_url_rule('/', 'index', _generate_index)
    app.add_url_rule('/index.html', 'index', _generate_index)
    app.add_url_rule('/<path:path>', 'deliver', _deliver)
  else:
    app.add_url_rule('/<path:path>', 'deliver', _deliver_production)
  app.add_url_rule('/api/buildInfo.json', 'build_info', _build_info)
  app.add_url_rule('/health', 'health', _health)
  return app



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################
from gevent import monkey
monkey.patch_all()


import logging.config  # noqa: E402
import logging  # noqa: E402
from . import config  # noqa: E402


# set configured registry
def _get_config():
  # check initialization
  if config._c is None:
    config._initialize()
  return config.view('phovea_server')


# added for testing
def _get_config_hdf():
  # check initialization
  if config._c is None:
    config._initialize()
  return config.view('phovea_data_hdf')

cc = _get_config()

cc_hdf = _get_config_hdf()

# configure logging
logging.config.dictConfig(cc.logging)
_log = logging.getLogger(__name__)


def enable_dev_mode():
  _log.info('enabling development mode')
  cc.set('env', 'development')
  cc.set('debug', True)
  cc.set('error_stack_trace', True)
  cc.set('nocache', False)


def enable_prod_mode():
  _log.info('enabling production mode')
  cc.set('env', 'production')
  cc.set('debug', False)
  cc.set('error_stack_trace', False)
  cc.set('nocache', False)


def _config_files():
  """
  list all known config files
  :return:
  """
  from .plugin import plugins
  return [p for p in (p.config_file() for p in plugins()) if p is not None]


def _resolve_launcher(launcher):
  """
  resolves the launcher if it is a string it will load the module and determine the function to use
  :param launcher:
  :return: launcher function
  """
  import six
  import os
  import importlib

  if isinstance(launcher, six.string_types):
    module_name, function_name = os.path.splitext(launcher)
    m = importlib.import_module(module_name)
    return m[function_name]

  return launcher


def set_default_subparser(parser, name, args=None):
  """default subparser selection. Call after setup, just before parse_args()
  name: is the name of the subparser to call by default
  args: if set is the argument list handed to parse_args()
  see https://stackoverflow.com/a/26378414

  , tested with 2.7, 3.2, 3.3, 3.4
  it works with 2.6 assuming argparse is installed
  """
  import sys
  import argparse

  subparser_found = False
  for arg in sys.argv[1:]:
    if arg in ['-h', '--help']:  # global help if no subparser
      break
  else:
    for x in parser._subparsers._actions:
      if not isinstance(x, argparse._SubParsersAction):
        continue
      for sp_name in list(x._name_parser_map.keys()):
        if sp_name in sys.argv[1:]:
          subparser_found = True
    if not subparser_found:
      # insert default in first position, this implies no
      # global options without a sub_parsers specified
      if args is None:
        sys.argv.insert(1, name)
      else:
        args.insert(0, name)


def _resolve_commands(parser):
  """
  Resolve commands from the phovea registy, loads the phovea extension, adds the instance to the command parser.
  """
  from .plugin import list as list_plugins

  # create a subparser
  subparsers = parser.add_subparsers(dest='cmd')

  default_command = None

  for command in list_plugins('command'):
    _log.info('add command ' + command.id)
    if hasattr(command, 'isDefault') and command.isDefault:
      default_command = command.id

    # create a argument parser for the command
    cmdparser = subparsers.add_parser(command.id)
    _log.info('loading and initializing the command: ' + command.id)

    # use the phovea extension point loading mechanism.
    # pass the parser as argument to the factory method so that the extension point (i.e., command)
    # can add further arguments to the parser (e.g., the address or port of the server).
    # the factory must return a launcher function, which gets the previously defined parser arguments as parameter.
    instance = command.load().factory(cmdparser)

    # register the instance as argument `launcher` and the command as `launcherid` to the command parser
    _log.info('add command instance to parser')
    cmdparser.set_defaults(launcher=instance, launcherid=command.id)

  return default_command


def _set_runtime_infos(args):
  """
  Set run time information, such as the executed command (registered as phovea extension point).
  The information is, for instance, used in the plugin.py when initializing the phovea registry.
  Additionally the configuration value `absoluteDir` is set.
  """
  import os
  runtime = cc.view('_runtime')
  runtime.set('command', args.launcherid)
  runtime.set('reloader', args.use_reloader)
  cc.set('absoluteDir', os.path.abspath(cc.get('dir')) + '/')


def run():
  """
  Run an application. The execution of the application can be configured using a command and arguments.
  Example terminal command:
  ```
  cd <workspace>
  python phovea_server --use_reloader --env dev api
  ```
  Supported arguments:
  `--use_reloader`: whether to automatically reload the server
  `--env`: environment mode (dev or prod)
  The last argument (e.g., `api`) is the command that must be registered as extension in the __init__.py and points to an execution file.
  Example:
  ```py
  registry.append('command', 'api', 'phovea_server.server', {'isDefault': True})
  ```
  The example registers the api command that runs the `create()` factory method from the server.py.
  """
  import argparse

  parser = argparse.ArgumentParser(description='Phovea Server')
  parser.add_argument('--use_reloader', action='store_true', help='whether to automatically reload the server')
  parser.add_argument('--env', default=cc.get('env'), help='environment mode (dev or prod)')

  # parse before to enable correct plugin discovery
  args = parser.parse_known_args()[0]
  if args.env.startswith('dev'):
    enable_dev_mode()
  else:
    enable_prod_mode()

  # resolve the default command to decide which application to launch
  default_command = _resolve_commands(parser)
  if default_command is not None:
    # set a default subparse to extract the defined arguments from the instance to the main arguments (?)
    set_default_subparser(parser, default_command)

  args = parser.parse_args()

  _set_runtime_infos(args)

  main = args.launcher(args)  # execute the launcher function, which returns another function

  if args.use_reloader:
    _log.info('start application using reloader...')
    run_with_reloader(main, extra_files=_config_files())
  else:
    _log.info('start application...')
    main()


def create_embedded():
  """
  Imports the phovea_server and creates an application
  """
  from .server import create_application
  return create_application()


# copied code of method run_with_reloader from werkzeug._reloader, because it causes import problems otherwise
def run_with_reloader(main_func, extra_files=None, interval=1, reloader_type="auto"):
  """Run the given function in an independent python interpreter."""
  import signal
  import os
  import sys
  import threading
  from werkzeug._reloader import reloader_loops, ensure_echo_on

  reloader = reloader_loops[reloader_type](extra_files, interval)
  signal.signal(signal.SIGTERM, lambda *args: sys.exit(0))
  try:
    if os.environ.get("WERKZEUG_RUN_MAIN") == "true":
      ensure_echo_on()
      t = threading.Thread(target=main_func, args=())
      t.setDaemon(True)
      t.start()
      reloader.run()
    else:
      sys.exit(reloader.restart_with_reloader())
  except KeyboardInterrupt:
    pass



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################

from builtins import map
from builtins import str
from builtins import object
from ._utils import replace_variables
from .config import view, _c, _initialize
import logging

# check initialization
if _c is None:
  _initialize()
cc = view('phovea_server')
_log = logging.getLogger(__name__)


def is_disabled_plugin(p):
  import re

  def check(disable):
    return isinstance(disable, str) and re.match(disable, p.id)

  return any(map(check, cc.disable['plugins']))


def is_development_mode():
  return cc.env.startswith('dev')


def is_disabled_extension(extension, extension_type, p):
  import re

  if is_disabled_plugin(p):
    return True

  def check_elem(k, v):
    vk = extension_type if k == 'type' else extension[k]
    return re.match(v, vk)

  def check(disable):
    if isinstance(disable, str):
      return re.match(disable, extension['id'])
    return all(check_elem(k, v) for k, v in disable.items())

  return any(map(check, cc.disable['extensions']))


def _resolve_server_config(d, vars={}):
  import six
  if isinstance(d, six.string_types):  # not a string
    return replace_variables(d, vars)
  elif isinstance(d, list):
    return [_resolve_server_config(i) for i in d]
  elif isinstance(d, dict):
    for k, v in d.items():
      d[k] = _resolve_server_config(v)
    return d
  return d


# extend a dictionary recursivly
def _extend(target, w):
  for k, v in w.items():
    if isinstance(v, dict):
      if k not in target:
        target[k] = _extend({}, v)
      else:
        target[k] = _extend(target[k], v)
    else:
      target[k] = v
  return target


def _git_head(cwd):
  import subprocess
  try:
    output = subprocess.check_output(['git', 'rev-parse', '--verify', 'HEAD'], cwd=cwd)
    return output.strip()
  except subprocess.CalledProcessError:
    return 'error'


def _resolve_plugin(p):
  import os.path
  if os.path.isdir(os.path.join(p.folder, '.git')) and p.repository:
    repo = p.repository
    if repo.endswith('.git'):
      repo = repo[0:-4]
    return str(repo) + '/commit/' + str(_git_head(p.folder))
  # not a git repo
  return p.version


class DirectoryPlugin(object):
  def __init__(self, package_file):
    import json
    import os.path as p
    folder = p.dirname(package_file)
    with open(package_file) as f:
      pkg = json.load(f)
    self.id = pkg['name']
    self._clean_id = self.id.lower().replace('-', '_')
    self.pkg = pkg
    self.name = self.id
    desc = pkg.get('description', '').split('\n')
    self.title = desc.pop(0) if len(desc) > 1 else self.name
    self.description = '\n'.join(desc)
    self.homepage = pkg.get('homepage')
    self.version = pkg['version']
    self.extensions = []
    self.repository = pkg.get('repository', {}).get('url')
    self.folder = folder

  def is_app(self):
    import os.path as p
    f = p.join(self.folder, 'build', 'index.html')
    return p.exists(f)

  def config_file(self):
    import os.path as p
    for f in [p.join(self.folder, 'config.json'), p.join(self.folder, self.id, 'config.json'), p.join(self.folder, self._clean_id, 'config.json')]:
      if p.exists(f):
        return f
    return None

  def register(self, reg):
    import os.path as p
    import sys

    def regfile(f):
      if not p.exists(f):
        return False
      # append path ../__init__.py
      sys.path.insert(0, p.abspath(p.dirname(p.dirname(f))))
      import importlib
      module = p.basename(p.dirname(f))
      _log.info('importing module: %s', module)
      m = importlib.import_module(module)
      if hasattr(m, 'phovea'):
        m.phovea(reg)
      return True

    regfile(p.join(self.folder, '__init__.py'))
    if not regfile(p.join(self.folder, self.id, '__init__.py')):
      regfile(p.join(self.folder, self._clean_id, '__init__.py'))

  @property
  def resolved(self):
    return _resolve_plugin(self)


class DirectoryProductionPlugin(object):
  def __init__(self, folder):
    import os.path as p
    import json
    self.folder = folder
    self.extensions = []
    if p.exists(p.join(folder, 'buildInfo.json')):
      with open(p.join(folder, 'buildInfo.json')) as f:
        pkg = json.load(f)
      self.id = pkg['name']
      self.name = self.id
      desc = pkg.get('description', '').split('\n')
      self.title = desc.pop(0) if len(desc) > 1 else self.name
      self.description = '\n'.join(desc)
      self.homepage = pkg.get('homepage')
      self.version = pkg['version']
      self.repository = pkg.get('repository', '')
    else:
      self.id = p.basename(folder)
      self.name = self.id
      self.title = self.id
      self.description = ''
      self.homepage = ''
      self.version = ''
      self.repository = ''

  @staticmethod
  def is_app():
    return False

  def config_file(self):
    import os.path as p
    f = p.join(self.folder, 'config.json')
    return f if p.exists(f) else None

  def register(self, reg):
    import os.path as p
    import sys

    def regfile(f):
      if not p.exists(f):
        return
      # append path ../__init__.py
      sys.path.insert(0, p.abspath(p.dirname(p.dirname(f))))
      import importlib
      module = p.basename(p.dirname(f))
      _log.info('importing module: %s', module)
      m = importlib.import_module(module)
      if hasattr(m, 'phovea'):
        m.phovea(reg)

    regfile(p.join(self.folder, '__init__.py'))

  @property
  def resolved(self):
    return self.version


class EntryPointPlugin(object):
  def __init__(self, entry_point, config_entry_point):
    import os.path as p
    import json
    self.id = entry_point.name
    self.name = self.id
    self.title = self.name
    self.description = ''
    self.version = ''
    self.extensions = []
    self.repository = None
    self._loader = entry_point.load
    self._config = config_entry_point.load

    # guess folder
    f = self.config_file()
    import os.path
    self.folder = os.path.dirname(f) if f else '.'

    if p.exists(p.join(self.folder, 'buildInfo.json')):
      with open(p.join(self.folder, 'buildInfo.json')) as f:
        pkg = json.load(f)
      desc = pkg.get('description', '').split('\n')
      self.title = desc.pop(0) if len(desc) > 1 else self.name
      self.description = '\n'.join(desc)
      self.homepage = pkg.get('homepage')
      self.version = pkg['version']
      self.repository = pkg.get('repository', '')

  @staticmethod
  def is_app():
    return False

  def config_file(self):
    if self._config is not None:
      return self._config()()
    return None

  def register(self, reg):
    self._loader(reg)

  @property
  def resolved(self):
    return self.version


class RegHelper(object):
  def __init__(self, plugin):
    self._items = []
    self._plugin = plugin

  def __iter__(self):
    return iter(self._items)

  def append(self, type_, id_, module_, desc=None):
    desc = {} if desc is None else desc
    desc['type'] = type_
    desc['id'] = id_
    desc['module'] = module_
    desc['plugin'] = self._plugin
    self._items.append(desc)


def _find_entry_point_plugins():
  import pkg_resources as p
  configs = {ep.name: ep for ep in p.iter_entry_points(group='phovea.config')}
  return [EntryPointPlugin(ep, configs.get(ep.name)) for ep in p.iter_entry_points(group='phovea.registry')]


def _find_development_neighbor_plugins():
  import os.path as p
  import glob
  import itertools
  prefix = ['./', '../', '../../']
  suffix = ['', 'p/', 'public/']
  files = []
  for pre, s in itertools.product(prefix, suffix):
    files.extend((p.abspath(pi) for pi in glob.glob(pre + s + '*/package.json')))
  # files contains all plugins
  return [DirectoryPlugin(pi) for pi in files]


def _find_production_neighbor_plugins():
  import os.path as p
  import glob
  base_dir = p.dirname(p.dirname(__file__))
  # all dirs having both __init__.py and config.json contained
  dirs = [p.dirname(p.abspath(pi)) for pi in glob.glob(base_dir + '/*/__init__.py')]
  dirs = [d for d in dirs if p.exists(p.join(d, 'config.json'))]
  # files contains all plugins
  v = [DirectoryProductionPlugin(d) for d in dirs]
  return v


class PluginMetaData(object):
  def __init__(self):
    self.plugins = []

    if not is_development_mode():
      entrypoints = _find_entry_point_plugins()
      self.plugins.extend(p for p in entrypoints if not is_disabled_plugin(p))

    entrypoint_ids = frozenset([p.id for p in self.plugins])

    if is_development_mode():
      _log.info('looking for development neighbors')
      neigbhors = _find_development_neighbor_plugins()
    else:
      _log.info('looking for production neighbors')
      neigbhors = _find_production_neighbor_plugins()

    self.plugins.extend(p for p in neigbhors if not is_disabled_plugin(p) and p.id not in entrypoint_ids)

    self.plugins.sort(key=lambda p: p.id)

    _log.info('discovered %d plugins: %s', len(self.plugins), [d.id for d in self.plugins])

    self.server_extensions = []
    for p in self.plugins:
      reg = RegHelper(p)
      p.register(reg)
      ext = [r for r in reg if not is_disabled_extension(r, 'python', p)]
      p.extensions = ext
      self.server_extensions.extend(ext)


def parse():
  p = PluginMetaData()
  return p



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################

from . import launch

application = launch.create_embedded()



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object
from .dataset_def import ADataSetEntry
import abc


class GraphNode(object):
  def __init__(self, t, id, attrs=None):
    self.type = t
    self.id = id
    self.attrs = {} if attrs is None else attrs

  def asjson(self):
    return dict(type=self.type, id=self.id, attrs=self.attrs)


class GraphEdge(object):
  def __init__(self, t, id, source=None, target=None, attrs=None):
    self.type = t
    self.id = id
    self.source = source
    self.target = target
    self.attrs = {} if attrs is None else attrs

  def asjson(self):
    return dict(type=self.type,
                id=self.id,
                source=self.source,
                target=self.target,
                attrs=self.attrs)


class AGraph(ADataSetEntry, metaclass=abc.ABCMeta):
  def __init__(self, name, project, id=None, attrs=None):
    super(AGraph, self).__init__(name, project, 'graph', id)
    self.attrs = {} if attrs is None else attrs

  @abc.abstractmethod
  def nodes(self, range=None):
    return []

  @property
  def nnodes(self):
    return len(self.nodes())

  @abc.abstractmethod
  def edges(self, range=None):
    return []

  @property
  def nedges(self):
    return len(self.edges())

  def to_description(self):
    r = super(AGraph, self).to_description()
    r['size'] = [self.nnodes, self.nedges]
    r['attrs'] = self.attrs
    return r

  def asjson(self, range=None):
    nodes = [a.asjson() for a in self.nodes(None if range is None else range[0])]
    edges = [a.asjson() for a in self.edges(None if range is None else range[1])]

    r = dict(nodes=nodes, edges=edges)
    return r

  def add_node(self, data):
    return False

  def update_node(self, data):
    return False

  def get_node(self, id):
    return next((n for n in self.nodes() if n.id == id), None)

  def remove_node(self, id):
    return False

  def add_edge(self, data):
    return False

  def get_edge(self, id):
    return next((n for n in self.edges() if n.id == id), None)

  def update_edge(self, data):
    return False

  def remove_edge(self, id):
    return False

  def clear(self):
    return False

  def incoming_edges(self, node):
    return (e for e in self.edges() if e.target == node.id)

  def outgoing_edges(self, node):
    return (e for e in self.edges() if e.source == node.id)

  def resolve_edges(self, edges):
    to_find = set()
    edges = list(edges)
    for e in edges:
      to_find.add(e.source)
      to_find.add(e.target)
    n = {n.id: n for n in self.nodes() if n.id in to_find}
    return ((e, n[e.source], n[e.target]) for e in edges)


def _resolve_parser(format):
  from .plugin import list as list_plugins
  for p in list_plugins('graph-parser'):
    if p.format == format:
      return p.load()


def _guess_format(format_field, files):
  if format_field:
    return format_field

  if len(files) > 0:
    # use the file extension as a hint
    import os.path
    fn = files[0].filename
    name, ext = os.path.splitext(fn)
    return ext.lower()
  return 'json'  # default


def parse(args, files):
  format = _guess_format(args.get('format', None), files)
  formatter = _resolve_parser(format)
  if formatter:
    return formatter.factory(args, files)
  return None



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str
from builtins import object
from . import plugin as p
import sys

ANONYMOUS = 'anonymous'


class User(object):
  def __init__(self, id):
    self.id = id
    self.name = ANONYMOUS
    self.roles = [ANONYMOUS]

  def get_id(self):
    return str(self.id)

  @property
  def is_authenticated(self):
    return False

  @property
  def is_active(self):
    return False

  @property
  def is_anonymous(self):
    return self.name == ANONYMOUS

  def has_role(self, role):
    return role in self.roles

  def __eq__(self, other):
    """
    Checks the equality of two `UserMixin` objects using `get_id`.
    """
    if isinstance(other, User):
      return self.get_id() == other.get_id()
    return NotImplemented

  def __ne__(self, other):
    """
    Checks the inequality of two `UserMixin` objects using `get_id`.
    """
    equal = self.__eq__(other)
    if equal is NotImplemented:
      return NotImplemented
    return not equal

  if sys.version_info[0] != 2:  # pragma: no cover
    # Python 3 implicitly set __hash__ to None if we override __eq__
    # We set it back to its default implementation
    __hash__ = object.__hash__


ANONYMOUS_USER = User(ANONYMOUS)


class SecurityManager(object):
  """
  a basic security manager
  """

  def __init__(self):
    pass

  def login_required(self, f):
    return f

  def login(self, username, extra_fields={}):
    """logs the given user in
    :returns the logged in user object or None if login failed
    """
    return User('')

  def logout(self):
    """
    logs the current logged in user out
    """
    pass

  @property
  def current_user(self):
    """
    :returns the current logged in user
    """
    return User('')

  def is_authenticated(self):
    """whether the current user is authenticated
    """
    return self.current_user.is_authenticated

  def has_role(self, role):
    """whether the current use has the role
    """
    return self.current_user.has_role(role)

  def init_app(self, app):
    pass

  def add_login_routes(self, app):
    pass


class DummyManager(SecurityManager):
  """
  a dummy implementation of the security manager where everyone is authenticated
  """

  def is_authenticated(self):
    return True

  def has_role(self, role):
    return True


_manager = None


def manager():
  """
  :return: the security manager
  """
  global _manager
  if _manager is None:
    _manager = p.lookup('security_manager')
    if _manager is None:
      _manager = DummyManager()
  return _manager


def is_logged_in():
  return manager().is_authenticated()


def current_username():
  u = manager().current_user
  return u.name if hasattr(u, 'name') else ANONYMOUS


def current_user():
  user = manager().current_user
  if user.is_anonymous:
    return ANONYMOUS_USER
  return user


def login_required(f):
  """
  Decorator for views that require login.
  """
  return manager().login_required(f)


def init_app(app):
  """
  initializes this app by for login mechanism
  :param app:
  :return:
  """
  manager().init_app(app)


def add_login_routes(app):
  """
  initializes this flask for providing access to /login and /logout
  :param app:
  :return:
  """
  manager().add_login_routes(app)


PERMISSION_READ = 4
PERMISSION_WRITE = 2
PERMISSION_EXECUTE = 1


def to_number(p_set):
  return (PERMISSION_READ if PERMISSION_READ in p_set else 0) + \
         (PERMISSION_WRITE if PERMISSION_WRITE in p_set else 0) + \
         (PERMISSION_EXECUTE if PERMISSION_EXECUTE in p_set else 0)


def to_string(p_set):
  return ('r' if PERMISSION_READ in p_set else '-') + \
         ('w' if PERMISSION_WRITE in p_set else '-') + \
         ('x' if PERMISSION_EXECUTE in p_set else '-')


def _from_number(p):
  r = set()
  if p >= 4:
    r.add(PERMISSION_READ)
    p -= 4
  if p >= 2:
    r.add(PERMISSION_WRITE)
    p -= 2
  if p >= 1:
    r.add(PERMISSION_EXECUTE)
  return r


DEFAULT_PERMISSION = 744


def _decode(permission=DEFAULT_PERMISSION):
  permission = int(permission)
  others = _from_number(permission % 10)
  group = _from_number((permission // 10) % 10)
  user = _from_number((permission // 100) % 10)
  buddies = _from_number((permission // 1000) % 10)
  return user, group, others, buddies


def _is_equal(a, b):
  if a == b:
    return True
  if not a or not b:
    return False
  a = a.lower()
  b = b.lower()
  return a == b


def _includes(items, item):
  if not item:
    return False
  for check in items:
    if _is_equal(check, item):
      return True
  return False


def can(item, permission, user=None):
  if user is None:
    user = current_user()

  if not isinstance(item, dict):
    # assume we have an object
    item = {
      'creator': getattr(item, 'creator', ANONYMOUS),
      'buddies': getattr(item, 'buddies', []),
      'group': getattr(item, 'group', ANONYMOUS),
      'permissions': getattr(item, 'permissions', DEFAULT_PERMISSION)
    }

  owner, group, others, buddies = _decode(item.get('permissions', DEFAULT_PERMISSION))

  # I'm the creator
  if _is_equal(user.name, item.get('creator', ANONYMOUS)) and permission in owner:
    return True

  # check if I'm in the buddies list
  if 'buddies' in item and _includes(item.get('buddies'), user.name) and permission in buddies:
    return True

  # check if I'm in the group
  if 'group' in item and _includes(user.roles, item.get('group')) and permission in group:
    return True

  return permission in others


def can_read(data_description, user=None):
  return can(data_description, PERMISSION_READ, user)


def can_write(data_description, user=None):
  return can(data_description, PERMISSION_WRITE, user)


def can_execute(data_description, user=None):
  return can(data_description, PERMISSION_EXECUTE, user)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str
import logging

_log = logging.getLogger(__name__)


# extend a dictionary recursively
def extend(target, w):
  for k, v in w.items():
    if isinstance(v, dict):
      if k not in target:
        target[k] = extend({}, v)
      else:
        target[k] = extend(target[k], v)
    else:
      target[k] = v
  return target


def replace_variables_f(s, lookup):
  import re
  s = str(s)
  if re.match(r'^\$\{([^}]+)\}$', s):  # full string is a pattern
    s = s[2:len(s) - 1]
    v = lookup(s)
    if v is None:
      _log.error('cant resolve ' + s)
      return '$unresolved$'
    return v

  def match(m):
    v = lookup(m.group(1))
    if v is None:
      _log.error('cant resolve ' + m.group(1))
      return '$unresolved$'
    return str(v)

  return re.sub(r'\$\{([^}]+)\}', match, s)


def replace_variables(s, variables):
  return replace_variables_f(s, lambda x: variables.get(x, None))


def replace_nested_variables(obj, lookup):
  if isinstance(obj, list):
    return [replace_nested_variables(o, lookup) for o in obj]
  elif isinstance(obj, str):
    return replace_variables_f(obj, lookup)
  elif isinstance(obj, dict):
    return {k: replace_nested_variables(v, lookup) for k, v in obj.items()}
  return obj



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################

import logging.config
import threading
import time


# set configured registry
def _get_config():
  from . import config
  return config.view('phovea_server')


# append the plugin directories as primary lookup path
cc = _get_config()
# configure logging
logging.config.dictConfig(cc.logging)
_log = logging.getLogger(__name__)


def _add_no_cache_header(response):
  """
  disable caching on the response
  :param response:
  :return:
  """
  #  response.headers['Last-Modified'] = datetime.now()
  response.headers['Cache-Control'] = 'no-store, no-cache, must-revalidate, post-check=0, pre-check=0, max-age=0'
  response.headers['Pragma'] = 'no-cache'
  response.headers['Expires'] = '-1'
  return response


def _rest_cache(namespace):
  import os
  from os import path
  import hashlib
  from werkzeug import url_quote

  dir_name = path.join(cc.restCacheDir, namespace[1:])
  if not path.exists(dir_name):
    os.makedirs(dir_name)

  def to_filename(request):
    secure = url_quote(request.full_path[1:].replace('/', ' ').replace('?', '_').replace('=', '-').replace('&', '_'), safe=' ')
    key = hashlib.sha256(namespace + secure).hexdigest()
    file_name_old = '{}_{}.json'.format(secure[:min(128, len(secure))], key)
    file_name = '{}_{}.json'.format(secure[:min(64, len(secure))], key)  # use a shorter filename for restCache, to avoid file system errors with too long file names
    return {
      "file_name_128": file_name_old.replace('%', '_'),  # file_name_128 = file name with 128 characters in front of the hash
      "file_name_64": file_name.replace('%', '_')  # file_name_64 = file name with 64 characters in front of the hash
    }

  def save(response):
    from flask import request

    if request.method != 'GET' or response.status_code != 200 or response.mimetype != 'application/json' or response.is_streamed or response.cache_control.no_cache:
      return response

    file_name = to_filename(request)['file_name_64']  # only use the short file name for storing new files

    _log.info('cache %s %s -> %s %s', namespace, request.full_path, dir_name, file_name)

    with open(path.join(dir_name, file_name), 'w') as f:
      f.write(response.get_data())

    # print(response.mimetype, data)
    return response

  def load():
    from flask import request, send_from_directory

    file_names = to_filename(request)

    # use the legacy file name ('file_name_128') if a file with this name exists and the new file name format otherwise
    file_name = file_names['file_name_128'] if path.exists(file_names['file_name_128']) else file_names['file_name_64']

    full = path.join(dir_name, file_name)

    if path.exists(full):
      _log.info('use cache cache %s %s <- %s', namespace, request.full_path, full)
      return send_from_directory(path.abspath(dir_name), file_name, add_etags=False)
    return None

  return save, load


def _to_stack_trace():
  import traceback
  return traceback.format_exc()


def _exception_handler(error):
  return 'Internal Server Error\n' + _to_stack_trace(), 500


def _init_app(app, namespace, is_default_app=False):
  """
  initializes an application by setting common properties and options
  :param app:
  :param is_default_app:
  :return:
  """
  from . import security

  if hasattr(app, 'got_first_request') and app.got_first_request:
    _log.warn('already inited: ' + str(app))
    return

  _log.info('init application: ' + str(app))

  if hasattr(app, 'debug'):
    app.debug = cc.debug
  if cc.nocache and hasattr(app, 'after_request'):
    app.after_request(_add_no_cache_header)
  if cc.error_stack_trace and hasattr(app, 'register_error_handler'):
    app.register_error_handler(500, _exception_handler)

  if cc['restCache'] == 'record' and hasattr(app, 'after_request'):
    app.after_request(_rest_cache(namespace)[0])
  if cc['restCache'] is True and hasattr(app, 'before_request'):
    app.before_request(_rest_cache(namespace)[1])

  if cc.secret_key:
    app.config['SECRET_KEY'] = cc.secret_key

  if cc.max_file_size:
    app.config['MAX_CONTENT_LENGTH'] = cc.max_file_size

  security.init_app(app)
  if is_default_app:
    security.add_login_routes(app)


# helper to plugin in function scope
def _loader(p):
  _log.info('add application: ' + p.id + ' at namespace: ' + p.namespace)

  def load_app():
    app = p.load().factory()
    _init_app(app, p.namespace)
    return app

  return load_app


def _pre_load_caches():
  """
  preload some manager to start them up
  :return:
  """
  c = _get_config().coldstart

  if c['assigner']:
    from .dataset import get_idmanager
    _log.info('initialize id assigner')
    get_idmanager()

  if c['mapping']:
    from .dataset import get_mappingmanager
    _log.info('initialize mapping manager')
    get_mappingmanager()


def create_application():
  from . import dispatcher
  from . import mainapp
  from .plugin import list as list_plugins
  from werkzeug.middleware.proxy_fix import ProxyFix

  # create a path dispatcher
  _default_app = mainapp.default_app()
  _init_app(_default_app, '/', True)
  _applications = {p.namespace: _loader(p) for p in list_plugins('namespace')}

  # create a dispatcher for all the applications
  application = dispatcher.PathDispatcher(_default_app, _applications)

  _pre_load_caches()

  return ProxyFix(application)


def create(parser):
  """
  Add arguments to the parser and return a launcher function, which in-turn creates a server instance
  parser: ArgumentParser that allows to add custom arguments. The arguments will be passed as parameter to the launcher function.
  """
  parser.add_argument('--port', '-p', type=int, default=cc.getint('port'),  # get default value from config.json
                      help='server port')
  parser.add_argument('--address', '-a', default=cc.get('address'),  # get default value from config.json
                      help='server address')
  parser.add_argument('--certfile', '-c', default=cc.get('certfile'),  # get default value from config.json
                      help='ssl certificate')
  parser.add_argument('--keyfile', '-k', default=cc.get('keyfile'),  # get default value from config.json
                      help='keyfile for ssl certificate')

  def _launcher(args):
    """
    Prepare the launch of the server instance
    args: contains the arguments that are parsed from the command line (or set in the config as default value)
    """
    from geventwebsocket.handler import WebSocketHandler
    from gevent.pywsgi import WSGIServer
    from flask.logging import default_handler
    import logging
    # from gevent import monkey

    # create phovea server application
    application = create_application()

    # add handler for wsgi's logger
    logger = logging.getLogger('wsgi')
    logger.setLevel(logging.INFO)
    _log.addHandler(default_handler)

    _log.info('prepare server that will listen on %s:%s [cert=%s, key=%s]', args.address, args.port, args.certfile, args.keyfile)
    # Test whether monkey.patch_all() has been used correctly,
    # keys have to be set
    # _log.warn(monkey.saved.keys())
    if args.certfile and args.keyfile:
      http_server = WSGIServer((args.address, args.port), application, keyfile=args.keyfile, certfile=args.certfile, handler_class=WebSocketHandler)
    else:
      http_server = WSGIServer((args.address, args.port), application, handler_class=WebSocketHandler)

    server = http_server.serve_forever  # return function name only; initialization will be done later

    # load `after_server_started` extension points which are run immediately after server started,
    # so all plugins should have been loaded at this point of time
    # the hooks are run in a separate (single) thread to not block the main execution of the server
    t = threading.Thread(target=_load_after_server_started_hooks)
    t.setDaemon(True)
    t.start()

    return server

  return _launcher


def _load_after_server_started_hooks():
  """
    Load and run all `after_server_started` extension points.
    The factory method of an extension implementing this extension point should return a function which is then executed here
  """
  from .plugin import list as list_plugins

  start = time.time()

  after_server_started_hooks = [p.load().factory() for p in list_plugins('after_server_started')]

  _log.info("Found %d `after_server_started` extension points to run", len(after_server_started_hooks))

  for hook in after_server_started_hooks:
    hook()

  _log.info("Elapsed time for server startup hooks: %d seconds", time.time() - start)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from .util import jsonify
from . import ns


def asrange(r=None):
  if r is None:
    return None
  from .range import parse
  return parse(r)


def _to_desc():
  if 'desc' in ns.request.values:
    import json
    n = json.loads(ns.request.values['desc'])
  else:
    n = ns.request.values
  return n


def format_json(dataset, range, args):
  d = dataset.asjson(range)
  if bool(args.get('f_pretty_print', False)):
    return jsonify(d, indent=' ')
  return jsonify(d)


def resolve_formatter(type, format):
  from .plugin import list as list_plugins
  for p in list_plugins(type + '-formatter'):
    if p.format == format:
      return p.load()
  formats = ','.join(p.format for p in list_plugins(type + '-formatter'))
  ns.abort(400, 'unknown format "{0}" possible formats are: {1}'.format(format, formats))


def _list_items(dataset_getter, name, datasetid):
  d = dataset_getter(datasetid, 'graph')
  if ns.request.method == 'GET':
    r = asrange(ns.request.args.get('range', None))
    return jsonify([n.asjson() for n in getattr(d, name + 's')(r[0] if r is not None else None)])

  if ns.request.method == 'DELETE':
    if not d.can_write():
      ns.abort(403)
    if d.clear():
      return jsonify(d.to_description(), indent=1)
    ns.abort(400)

  # post
  if not d.can_write():
    ns.abort(403)
  n = _to_desc()
  if getattr(d, 'add_' + name)(n):
    return jsonify(d.to_description(), indent=1)
  # invalid upload
  ns.abort(400)


def _handle_item(dataset_getter, name, datasetid, itemid):
  d = dataset_getter(datasetid, 'graph')
  if ns.request.method == 'GET':
    n = getattr(d, 'get_' + name)(itemid)
    return jsonify(n.asjson())

  if ns.request.method == 'DELETE':
    if not d.can_write():
      ns.abort(403)
    if getattr(d, 'remove_' + name)(itemid):
      return jsonify(d.to_description(), indent=1)
    ns.abort(400)

  # put
  if not d.can_write():
    ns.abort(403)
  n = _to_desc()
  n['id'] = itemid
  if getattr(d, 'update_' + name)(n):
    return jsonify(d.to_description(), indent=1)
  # invalid upload
  ns.abort(400)


def _list_type(dataset_getter, name='node'):
  def listitem(datasetid):
    return _list_items(dataset_getter, name, datasetid)

  def handleitem(datasetid, itemid):
    return _handle_item(dataset_getter, name, datasetid, itemid)

  return listitem, handleitem


def add_graph_handler(app, dataset_getter):
  @app.route('/graph/<datasetid>')
  @ns.etag
  def list_graphs(datasetid):
    d = dataset_getter(datasetid, 'graph')
    return jsonify(d.to_description())

  @app.route('/graph/<datasetid>/data')
  @ns.etag
  def get_graph_data(datasetid):
    d = dataset_getter(datasetid, 'graph')
    r = asrange(ns.request.args.get('range', None))
    formatter = resolve_formatter('graph', ns.request.args.get('format', 'json'))
    return formatter(d, r, args=ns.request.args)

  list_nodes, handle_node = _list_type(dataset_getter, 'node')
  app.add_url_rule('/graph/<datasetid>/node', 'list_nodes', ns.etag(list_nodes), methods=['GET', 'POST', 'DELETE'])
  app.add_url_rule('/graph/<datasetid>/node/<int:itemid>', 'handle_node', ns.etag(handle_node), methods=['GET', 'PUT', 'DELETE'])

  list_edges, handle_edge = _list_type(dataset_getter, 'edge')
  app.add_url_rule('/graph/<datasetid>/edge', 'list_edges', ns.etag(list_edges), methods=['GET', 'POST', 'DELETE'])
  app.add_url_rule('/graph/<datasetid>/edge/<int:itemid>', 'handle_edge', ns.etag(handle_edge), methods=['GET', 'PUT', 'DELETE'])

  # websocket = ws.Socket(app)
  # @websocket.route('/ws')
  # def graph_ws(socket):
  #  ws.websocket_loop(socket, dict(get_graph=(payload, s)))



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str
from . import ns, plugin, range
from .util import jsonify, to_json
import logging
from .dataset import list_idtypes, get_idmanager, iter, get_mappingmanager, get, list_datasets, add, remove
from flask import abort
from typing import List


app = ns.Namespace(__name__)
app_idtype = ns.Namespace(__name__)

_log = logging.getLogger(__name__)


@app.errorhandler(ValueError)
def on_value_error(error):
  _log.error('ValueError: (' + str(error) + ') at ' + str(ns.request.environ))
  _log.error(error)
  return '<strong>{2} - {0}</strong><pre>{1}</pre>'.format('ValueError', error, 500), 500


def _list_format_json(data):
  return jsonify(data)


def _list_format_treejson(data):
  r = dict()
  for d in data:
    levels = d['fqname'].split('/')
    act = r
    for level in levels[:-1]:
      if level not in act:
        act[level] = dict()
      act = act[level]
    act[d['name']] = d
  return jsonify(r, indent=1)


def _list_format_csv(data):
  delimiter = ns.request.args.get('f_delimiter', ';')

  def to_size(size):
    if size is None:
      return ''
    if isinstance(size, list):
      return ','.join(str(d) for d in size)
    return str(size)

  def gen():
    yield delimiter.join(['ID', 'Name', 'FQName', 'Type', 'Size', 'Entry'])
    for d in data:
      yield '\n'
      yield delimiter.join(
          [str(d['id']), d['name'], d['fqname'], d['type'], to_size(d.get('size', None)), to_json(d)])

  return ns.Response(gen(), mimetype='text/csv', headers={'Content-Disposition': 'attachment;filename=dataset.csv'})


def _to_query(query):
  keys = ['name', 'id', 'fqname', 'type']
  act_query = {k: v for k, v in query.items() if k in keys}
  if len(act_query) == 0:  # no query
    return lambda x: True
  import re

  def filter_elem(elem):
    return all((re.match(v, getattr(elem, k, '')) for k, v in act_query.items()))

  return filter_elem


@app.route('/', methods=['GET', 'POST'])
@ns.etag
def _list_datasets():
  if ns.request.method == 'GET':
    query = _to_query(ns.request.values)
    data = [d.to_description() for d in iter() if query(d)]

    limit = ns.request.values.get('limit', -1)
    if 0 < limit < len(data):
      data = data[:limit]

    format = ns.request.args.get('format', 'json')
    formats = dict(json=_list_format_json, treejson=_list_format_treejson, csv=_list_format_csv)
    if format not in formats:
      ns.abort(
          ns.make_response('invalid format: "{0}" possible ones: {1}'.format(format, ','.join(list(formats.keys()))), 400))
    return formats[format](data)
  else:
    return _upload_dataset(ns.request)


@app.route('/<dataset_id>', methods=['PUT', 'GET', 'DELETE', 'POST'])
@ns.etag
def _get_dataset(dataset_id):
  if ns.request.method == 'PUT':
    return _update_dataset(dataset_id, ns.request)
  elif ns.request.method == 'POST':
    return _modify_dataset(dataset_id, ns.request)
  elif ns.request.method == 'DELETE':
    return _remove_dataset(dataset_id)
  d = get(dataset_id)
  if d is None:
    return 'invalid dataset id "' + str(dataset_id) + '"', 404
  if not d.can_read():
    return 'not allowed', 403
  r = ns.request.args.get('range', None)
  if r is not None:
    r = range.parse(r)
  return jsonify(d.asjson(r))


@app.route('/<dataset_id>/desc')
@ns.etag
def _get_dataset_desc(dataset_id):
  d = get(dataset_id)
  if not d:
    return 'invalid dataset id "' + str(dataset_id) + '"', 404
  if not d.can_read():
    return 'not allowed', 403
  return jsonify(d.to_description())


def _dataset_getter(dataset_id, dataset_type):
  if isinstance(dataset_id, int) and dataset_id < 0:
    return [d for d in list_datasets() if d.type == dataset_type]
  t = get(dataset_id)
  if t is None:
    ns.abort(404, 'invalid dataset id "' + str(dataset_id) + '"')
  if t.type != dataset_type:
    ns.abort(400, 'the given dataset "' + str(dataset_id) + '" is not a ' + dataset_type)
  if not t.can_read():
    ns.abort(403, 'not allowed')
  return t


def _to_upload_desc(data_dict):
  if 'desc' in data_dict:
    import json
    return json.loads(data_dict['desc'])
  return data_dict


def _upload_dataset(request, id=None):
  try:
    # first choose the provider to handle the upload
    r = add(_to_upload_desc(request.values), request.files, id)
    if r:
      return jsonify(r.to_description(), indent=1)
    # invalid upload
    return 'invalid upload', 400
  except ValueError as e:
    return on_value_error(e)


def _update_dataset(dataset_id, request):
  try:
    old = get(dataset_id)
    if old is None:
      return _upload_dataset(request, dataset_id)
    if not old.can_write():
      return 'not allowed', 403
    r = old.update(_to_upload_desc(request.values), request.files)
    if r:
      return jsonify(old.to_description(), indent=1)
    # invalid upload
    return 'invalid upload', 400
  except ValueError as e:
    return on_value_error(e)


def _modify_dataset(dataset_id, request):
  try:
    old = get(dataset_id)
    if old is None:
      return 'invalid dataset id "' + str(dataset_id) + '"', 404
    if not old.can_write():
      return 'not allowed', 403
    r = old.modify(_to_upload_desc(request.values), request.files)
    if r:
      return jsonify(old.to_description(), indent=1)
      # invalid upload
    return 'invalid upload', 400
  except ValueError as e:
    return on_value_error(e)


def _remove_dataset(dataset_id):
  dataset = get(dataset_id)
  if dataset is None:
    return 'invalid dataset id "' + str(dataset_id) + '"', 404
  if not dataset.can_write():
    return 'not allowed', 403
  r = remove(dataset_id)
  if r:
    return jsonify(
        dict(state='success', msg='Successfully deleted dataset ' + dataset_id, id=dataset_id), indent=1)
  return 'invalid request', 400


def create_dataset():
  return app


@app_idtype.route('/')
@ns.etag
def _list_idtypes():
  return jsonify(list_idtypes())


@app_idtype.route('/<idtype>/map', methods=['GET', 'POST'])
def _map_ids(idtype):
  name = ns.request.values.get('id', None)
  if name is not None:
    return get_idmanager()([name], idtype)[0]
  names = ns.request.values.getlist('ids[]')
  return jsonify(get_idmanager()(names, idtype))


@app_idtype.route('/<idtype>/unmap', methods=['GET', 'POST'])
def _unmap_ids(idtype):
  name = ns.request.values.get('id', None)
  if name is not None:
    return get_idmanager().unmap([int(name)], idtype)[0]
  names = range.parse(ns.request.values.get('ids', ''))[0].tolist()
  return jsonify(get_idmanager().unmap(names, idtype))


@app_idtype.route('/<idtype>/')
@ns.etag
def _maps_to(idtype):
  mapper = get_mappingmanager()
  target_id_types = mapper.maps_to(idtype)
  return jsonify(target_id_types)


@app_idtype.route('/<idtype>/search')
def _search_ids(idtype):
  query = ns.request.args.get('q', None)
  max_results = int(ns.request.args.get('limit', 10))
  manager = get_idmanager()
  if query is None:
    return abort(400, 'Parameter "q" must be defined')
  if hasattr(manager, 'search'):
    return jsonify(manager.search(idtype, query, max_results))
  return jsonify([])


@app_idtype.route('/<idtype>/<to_idtype>', methods=['GET', 'POST'])
def _mapping_to(idtype, to_idtype):
  return _do_mapping(idtype, to_idtype, False)


@app_idtype.route('/<idtype>/<to_idtype>/search')
def _mapping_to_search(idtype, to_idtype):
  query = ns.request.args.get('q', None)
  max_results = int(ns.request.args.get('limit', 10))
  mapper = get_mappingmanager()
  if hasattr(mapper, 'search'):
    return jsonify(mapper.search(idtype, to_idtype, query, max_results))
  return jsonify([])


def _do_mapping(idtype, to_idtype, to_ids):
  mapper = get_mappingmanager()
  args = ns.request.values
  first_only = args.get('mode', 'all') == 'first'

  if 'id' in args:
    names = get_idmanager().unmap([int(args['id'])], idtype)
  elif 'ids' in args:
    names = get_idmanager().unmap(range.parse(args['ids'])[0].tolist(), idtype)
  elif 'q' in args:
    names = args['q'].split(',')
  elif 'q[]' in args:
    names = args.getlist('q[]')
  else:
    ns.abort(400)
    return

  # we'll call the mapper for each name individually to preserve the order of the names,
  # because otherwise it's not guaranteed that the results are in the same order as the names
  # and since we need to merge the results on the frontend occasionally we might mix them up
  mappers_list: List[List[List[str]]] = [mapper(idtype, to_idtype, [name]) for name in names]

  # flatten the list to be 2-dimensional since this is the format we used to return
  mapped_list: List[List[str]] = [mapping[0] for mapping in mappers_list]

  if first_only:
    mapped_list = [None if a is None or len(a) == 0 else a[0] for a in mapped_list]

  if to_ids:
    m = get_idmanager()
    if first_only:
      mapped_list = m(mapped_list, to_idtype)
    else:
      mapped_list = [m(entry, to_idtype) for entry in mapped_list]

  return jsonify(mapped_list)


@app_idtype.route('/<idtype>/<to_idtype>/map', methods=['GET', 'POST'])
def _mapping_to_id(idtype, to_idtype):
  return _do_mapping(idtype, to_idtype, True)


# add all specific handler
for handler in plugin.list('dataset-specific-handler'):
  p = handler.load()
  p(app, _dataset_getter)


def create_idtype():
  return app_idtype


if __name__ == '__main__':
  app.debug = True
  app.run(host='0.0.0.0', port=9000)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object, set
from .plugin import list as list_plugin
from itertools import chain
import logging

_log = logging.getLogger(__name__)


class MappingManager(object):
  """
  assigns ids to object using a redis database
  """
  def __init__(self, providers):
    self.mappers = {}
    self.paths = {}
    graph = {}
    for (from_idtype, to_idtype, mapper) in providers:
      # generate mapper mapping
      from_mappings = self.mappers.get(from_idtype, {})
      self.mappers[from_idtype] = from_mappings
      to_mappings = from_mappings.get(to_idtype, [])
      from_mappings[to_idtype] = to_mappings
      to_mappings.append(mapper)
      # generate type graph
      from_graph = graph.get(from_idtype, [])
      from_graph.append(to_idtype)
      graph[from_idtype] = from_graph
    # generate path map
    entries = self.known_idtypes()
    for _from in entries:
      # calculate all paths
      all_paths = {_to: self.__find_all_paths(graph, _from, _to) for _to in entries if _to != _from}
      # remove missing paths
      for key, value in list(all_paths.items()):
        if not value:
          del all_paths[key]
      self.paths[_from] = all_paths

  def known_idtypes(self):
    """
    returns a set of a all known id types in this mapping graph
    :return:
    """
    s = set()
    for from_, v in self.mappers.items():
      s.add(from_)
      for to_ in list(v.keys()):
        s.add(to_)
    return s

  def __find_all_paths(self, graph, start, end, path=[]):
    """
    Returns all possible paths in the graph from start to end
    :return: Array of all possible paths (string arrays) sorted by shortest path first
    """
    path = path + [start]
    if start == end:
      return [path]
    if start not in graph:
      return []
    paths = []
    for node in graph[start]:
      if node not in path:
        newpaths = self.__find_all_paths(graph, node, end, path)
        for newpath in newpaths:
          paths.append(newpath)
    return sorted(paths, key=len)

  def __resolve_single(self, from_idtype, to_idtype, ids):
    from_mappings = self.mappers.get(from_idtype, {})
    to_mappings = from_mappings.get(to_idtype, [])
    if not to_mappings:
      _log.warn('cannot find mapping from %s to %s', from_idtype, to_idtype)
      return [None for _ in ids]

    if len(to_mappings) == 1:
      # single mapping no need for merging
      return to_mappings[0](ids)

    # two way to preserve the order of the results
    r = [[] for _ in ids]
    rset = [set() for _ in ids]
    for mapper in to_mappings:
      mapped_ids = mapper(ids)
      for mapped_id, rlist, rhash in zip(mapped_ids, r, rset):
        for id in mapped_id:
          if id not in rhash:
            rlist.append(id)
            rhash.add(id)
    return r

  def merge_2d_arrays(self, source, lengths):
    """
    Merges the arrays of the source array according to the lengths array
    For example, [[1], [2], [3]] with lengths [2, 1] becomes [[1, 2], [3]]
    :return: Merged arrays
    """
    if(len(lengths) == 0 and len(source) == 0):
      return []
    assert len(lengths) > 0 and min(lengths) >= 1
    assert sum(lengths) == len(source)
    result = []
    i = 0
    for length in lengths:
      next_i = i + length
      result.append(source[i] if length == 1 else list(chain.from_iterable(source[i:next_i])))
      i = next_i
    return result

  def can_map(self, from_idtype, to_idtype):
    return self.paths.get(from_idtype, {}).get(to_idtype)

  def maps_to(self, from_idtype):
    return list(self.paths.get(from_idtype, {}).keys())

  def __call__(self, from_idtype, to_idtype, ids):
    # Get stored path instead of calculating them "on the fly"
    # paths = self.__find_all_paths(self.graph, from_idtype, to_idtype)
    paths = self.paths.get(from_idtype, {}).get(to_idtype)

    if not paths:
      _log.warn('Cannot find mapping from %s to %s', from_idtype, to_idtype)
      return [None for _ in ids]

    # Traverse shortest path only for now
    path = paths[0]
    if len(path) < 2:
      _log.warn('Invalid path given: %s', path)
      return [None for _ in ids]

    values = ids
    needs_merging = False
    lengths = []
    # Iterate over from, to tuples
    for i in range(1, len(path)):
      from_type = path[i-1]
      to_type = path[i]
      result = self.__resolve_single(from_type, to_type, values)

      # Check if needs merging
      if needs_merging:
        result = self.merge_2d_arrays(result, lengths)

      # Return on last iteration
      if i == len(path)-1:
        return result

      # Otherwise, check if every mapping was 1 to 1
      lengths = [len(x) for x in result]
      # If any result array is longer than 1, we need to flatten and later merge it
      needs_merging = max(lengths, default=0) > 1
      # Flatten result and assign to values
      values = list(chain.from_iterable(result))
    return result

  def search(self, from_idtype, to_idtype, query, max_results=None):
    """
    Searches for matches in the names of the given idtype.
    This operation does not resolve transitive mappings.
    :param query:
    :param max_results
    :return:
    """
    from_mappings = self.mappers.get(from_idtype, {})
    to_mappings = from_mappings.get(to_idtype, [])
    to_mappings = [m for m in to_mappings if hasattr(m, 'search')]

    if not to_mappings:
      _log.warn('cannot find mapping from %s to %s', from_idtype, to_idtype)
      return []

    if len(to_mappings) == 1:
      # single mapping no need for merging
      return to_mappings[0].search(query, max_results)

    rset = set()
    for mapper in to_mappings:
      results = mapper.search(query, max_results)
      for r in results:
        rset.add(r)
    return list(rset)


def create():
  # Load mapping providers
  providers = []
  for plugin in list_plugin('mapping_provider'):
    providers = providers + list(plugin.load().factory())
  return MappingManager(providers)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from flask import Flask
from flask_restplus import Api, apidoc, Resource  # noqa


def create_api(name, **kwargs):
  """
  utility function for creating an flask restplus api having the docu accessable under /doc and not the root
  """
  app = Flask(name)
  app.debug = True
  api = Api(app, ui=False, **kwargs)

  return app, api



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################
from logging.handlers import TimedRotatingFileHandler as Base


def phovea(registry):
  """
  extension points
  @param registry phovea registry
  register extension points
  :param registry:
  """
  registry.append('namespace', 'caleydo-dataset', 'phovea_server.dataset_api', {
      'namespace': '/api/dataset',
      'factory': 'create_dataset'
  })
  registry.append('namespace', 'caleydo-idtype', 'phovea_server.dataset_api', {
      'namespace': '/api/idtype',
      'factory': 'create_idtype'
  })
  registry.append('dataset-specific-handler', 'handler-table', 'phovea_server.dataset_specific', {
      'datatype': 'table',
      'factory': 'add_table_handler'
  })
  registry.append('dataset-specific-handler', 'handler-matrix', 'phovea_server.dataset_specific', {
      'datatype': 'matrix',
      'factory': 'add_matrix_handler'
  })
  registry.append('dataset-specific-handler', 'handler-vector', 'phovea_server.dataset_specific', {
      'datatype': 'vector',
      'factory': 'add_vector_handler'
  })
  registry.append('table-formatter', 'formatter-json', 'phovea_server.dataset_specific', {
      'format': 'json',
      'factory': 'format_json'
  })
  registry.append('table-formatter', 'formatter-csv', 'phovea_server.dataset_specific', {
      'format': 'csv',
      'factory': 'format_csv'
  })
  registry.append('matrix-formatter', 'formatter-json', 'phovea_server.dataset_specific', {
      'format': 'json',
      'factory': 'format_json'
  })
  registry.append('matrix-formatter', 'formatter-csv', 'phovea_server.dataset_specific', {
      'format': 'csv',
      'factory': 'format_csv'
  })
  registry.append('matrix-formatter', 'formatter-image', 'phovea_server.dataset_specific', {
      'format': 'png',
      'factory': 'format_image'
  })
  registry.append('matrix-formatter', 'formatter-image', 'phovea_server.dataset_specific', {
      'format': 'jpeg',
      'factory': 'format_image'
  })
  registry.append('matrix-formatter', 'formatter-image', 'phovea_server.dataset_specific', {
      'format': 'jpg',
      'factory': 'format_image'
  })
  registry.append('vector-formatter', 'formatter-json', 'phovea_server.dataset_specific', {
      'format': 'json',
      'factory': 'format_json'
  })
  registry.append('vector-formatter', 'formatter-csv', 'phovea_server.dataset_specific', {
      'format': 'csv',
      'factory': 'format_csv'
  })
  registry.append('dataset-provider', 'dataset-csv', 'phovea_server.dataset_csv')
  registry.append('json-encoder', 'numpy', 'phovea_server.json_encoder')
  registry.append('json-encoder', 'set-encoder', 'phovea_server.set_encoder', {})

  registry.append('manager', 'idmanager', 'phovea_server.assigner', dict(singleton=True))
  registry.append('manager', 'mappingmanager', 'phovea_server.mapper', dict(singleton=True))
  registry.append('mapping_provider', 'csv', 'phovea_server.mapping_csv')

  registry.append('dataset-specific-handler', 'handler-graph', 'phovea_server.graph_api', {
      'datatype': 'graph',
      'factory': 'add_graph_handler'
  })
  registry.append('graph-formatter', 'formatter-json', 'phovea_server.graph_api', {
      'format': 'json',
      'factory': 'format_json'
  })
  registry.append('graph-parser', 'parser-json', 'phovea_server.graph_parser', {
      'format': 'json',
      'factory': 'parse_json'
  })
  registry.append('command', 'api', 'phovea_server.server', {'isDefault': True})


def phovea_config():
  """
  :return: file pointer to config file
  """
  from os import path
  here = path.abspath(path.dirname(__file__))
  config_file = path.join(here, 'config.json')
  return config_file if path.exists(config_file) else None


def _ensure_exists(filename):
  import os
  # ensure directory exists
  d = os.path.dirname(filename)
  if not os.path.exists(d):
    os.makedirs(d)
  return filename


class TimedRotatingFileHandler(Base):
  def __init__(self, filename, *args, **kwargs):
    super(TimedRotatingFileHandler, self).__init__(_ensure_exists(filename), *args, **kwargs)



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str, object
import json
import io
import os
import csv
import numpy as np
from .dataset_def import ADataSetProvider, AColumn, AMatrix, AStratification, ATable, AVector
from . import config
from functools import cmp_to_key


def assign_ids(ids, idtype):
  from .plugin import lookup

  manager = lookup('idmanager')
  return np.array(manager(ids, idtype))


def fix_id(fqname):
  from .util import fix_id
  return fix_id(fqname)


def basic_description(data, type, path):
  import datetime
  from .security import current_username
  desc = dict(type=type,
              name=data.get('name', 'Uploaded File'),
              description=data.get('description', ''),
              creator=current_username,
              ts=datetime.datetime.utcnow(),
              path=os.path.basename(path))
  if 'group' in data:
    desc['group'] = data['group']
  if 'permissions' in data:
    desc['permissions'] = data['permissions']
  if 'buddies' in data:
    desc['buddies'] = data['buddies']

  return desc


class CSVEntryMixin(object):
  def __init__(self, desc, project):
    self._desc = desc
    folder = project.folder + '/data/' if not hasattr(project, 'inplace') else project.folder
    self._path = os.path.join(folder, self._desc['path'])
    del self._desc['path']
    self._project = project

    self._loaded = None

  def load(self):
    if self._loaded is not None:
      return self._loaded

    data = []
    with io.open(self._path, 'r', newline='', encoding=self._desc.get('encoding', 'utf-8')) as csvfile:
      reader = csv.reader(csvfile, delimiter=self._desc.get('separator', ','), quotechar=str(self._desc.get('quotechar', '|')))
      data.extend(reader)

    # print data
    def to_num(s):
      try:
        return float(s)  # for int, long and float
      except ValueError:
        return s

    header = data[0]
    data = [[to_num(v) if i > 0 else v for i, v in enumerate(row)] for row in data[1:]]
    data.insert(0, header)

    # convert to col, row and data
    self._loaded = self._process(data)
    return self._loaded

  def _process(self, data):
    return data

  def to_description(self):
    return self._desc

  def idtypes(self):
    return [v for k, v in self._desc.items() if k in ['rowtype', 'coltype', 'idtype']]


def guess_color(name, i):
  name = name.lower()
  colors = dict(name='blue', female='red', deceased='#e41a1b', living='#377eb8')
  if name in colors:
    return colors[name]
  l = ['#8dd3c7', '#ffffb3', '#bebada', '#fb8072', '#80b1d3', '#fdb462', '#b3de69', '#fccde5', '#d9d9d9', '#bc80bd',
       '#ccebc5', '#ffed6f']
  return l[i % len(l)]


def cmp_string(a, b):
  if a == b:
    return 0
  return -1 if a < b else +1


class CSVStratification(CSVEntryMixin, AStratification):
  def __init__(self, desc, project):
    AStratification.__init__(self, desc['name'], project.id, desc['type'], desc.get('id', None))
    CSVEntryMixin.__init__(self, desc, project)
    desc['fqname'] = self.fqname
    desc['id'] = self.id

    self.idtype = desc['idtype']
    for i, g in enumerate(desc['groups']):
      if 'color' not in g:
        g['color'] = guess_color(g['name'], i)

  def _process(self, data):
    def to_string(v):
      if type(v) is float:
        return str(int(v))
      return str(v)

    d = [dict(row=row[0], i=i, cluster=to_string(row[1])) for i, row in enumerate(data[1:])]

    groups = [str(g['name']) for g in self._desc['groups']]

    def cmp(a, b):
      ga = groups.index(a['cluster'])
      gb = groups.index(b['cluster'])
      if ga != gb:
        return ga - gb

      r = cmp_string(a['cluster'], b['cluster'])
      if r != 0:
        return r
      return cmp_string(a['row'], b['row']) if r == 0 else r

    d.sort(key=cmp_to_key(cmp))  # sort by cluster;
    clusters = dict()
    for di in d:
      c = di['cluster']
      if c in clusters:
        clusters[c].append(di['i'])
      else:
        clusters[c] = [di['i']]

    colors = {g['name']: g['color'] for g in self._desc['groups']}
    clusters = [dict(name=k, range=clusters.get(k, []), color=colors.get(k, 'gray')) for k in groups]

    rows = np.array([di[0] for di in data[1:]])
    return {'rows': rows,
            'rowIds': assign_ids(rows, self.idtype),
            'groups': clusters
            }

  def rows(self, range=None):
    n = self.load()['rows']
    if range is None:
      return n
    return n[range.asslice()]

  def rowids(self, range=None):
    n = self.load()['rowIds']
    if range is None:
      return n
    return n[range.asslice()]

  def groups(self):
    return self.load()['groups']

  def asjson(self, range=None):
    return self.load()

  @staticmethod
  def parse(data, path, project, id=None):
    desc = basic_description(data, 'stratification', path)
    desc['idtype'] = data.get('idtype', data.get('rowtype', 'unknown'))

    for k, v in data.items():
      if k not in desc:
        desc[k] = v
    if id is not None:
      desc['id'] = id
    if 'size0' in data and 'ngroups' in data:
      desc['size'] = [int(data['size0'])]
      del desc['size0']
      desc['ngroups'] = int(data['ngroups'])
    else:  # derive from the data
      clusters = set()
      count = 0
      with io.open(path, 'r', newline='', encoding=desc.get('encoding', 'utf-8')) as csvfile:
        reader = csv.reader(csvfile, delimiter=desc.get('separator', ','), quotechar=str(desc.get('quotechar', '|')))
        for row in reader:
          count += 1
          clusters.add(row[1])
      desc['size'] = [count]
      desc['ngroups'] = len(clusters)

    return CSVStratification(desc, project)


class CSVMatrix(CSVEntryMixin, AMatrix):
  def __init__(self, desc, project):
    AMatrix.__init__(self, desc['name'], project.id, desc['type'], desc.get('id', None))
    CSVEntryMixin.__init__(self, desc, project)
    desc['fqname'] = self.fqname
    desc['id'] = self.id
    self.rowtype = desc['rowtype']
    self.coltype = desc['coltype']
    self.value = desc['value']['type']
    self.range = desc['value']['range']
    self.shape = desc['size']

  def _process(self, data):
    cols = np.array(data[0][1:])
    rows = np.array([x[0] for x in data[1:]])
    is_number = self.value == 'real' or self.value == 'int'

    if is_number:
      vs = [[np.NaN if v == 'NA' or v == '' else v for v in x[1:]] for x in data[1:]]
      # import numpy.ma as ma
      # dd = ma.masked_equal(np.array(vs), np.NaN)
      dd = np.array(vs)
    else:
      dd = np.array([x[1:] for x in data[1:]])
    return {'cols': cols,
            'colIds': assign_ids(cols, self.coltype),
            'rows': rows,
            'rowIds': assign_ids(rows, self.rowtype),
            'data': dd
            }

  def rows(self, range=None):
    n = self.load()['rows']
    if range is None:
      return n
    return n[range.asslice()]

  def rowids(self, range=None):
    n = self.load()['rowIds']
    if range is None:
      return n
    return n[range.asslice()]

  def cols(self, range=None):
    n = self.load()['cols']
    if range is None:
      return n
    return n[range.asslice()]

  def colids(self, range=None):
    n = self.load()['colIds']
    if range is None:
      return n
    return n[range.asslice()]

  def asnumpy(self, range=None):
    n = self.load()['data']
    if range is None:
      return n

    rows = range[0].asslice()
    cols = range[1].asslice()
    d = None
    if isinstance(rows, list) and isinstance(cols, list):
      # fancy indexing in two dimension doesn't work
      d_help = n[rows, :]
      d = d_help[:, cols]
    else:
      d = n[rows, cols]

    if d.ndim == 1:
      # two options one row and n columns or the other way around
      if rows is Ellipsis or (isinstance(rows, list) and len(rows) > 1):
        d = d.reshape((d.shape[0], 1))
      else:
        d = d.reshape((1, d.shape[0]))
    elif d.ndim == 0:
      d = d.reshape((1, 1))
    return d

  @staticmethod
  def parse(data, path, project, id=None):
    desc = basic_description(data, 'matrix', path)
    desc['rowtype'] = data.get('rowtype', 'unknown')
    desc['coltype'] = data.get('coltype', 'unknown')
    desc['value'] = dict(type=data.get('value_type', 'real'))

    for k, v in data.items():
      if k not in desc:
        desc[k] = v
    if id is not None:
      desc['id'] = id

    if all((k in data) for k in ['size0', 'size1', 'value_min', 'value_max']):
      desc['size'] = [int(data['size0']), int(data['size1'])]
      del desc['size0']
      del desc['size1']
      desc['value']['range'] = [float(data['value_min']), float(data['value_max'])]
      del desc['value_min']
      del desc['value_max']
    else:  # derive from the data
      rows = 0
      cols = None
      min_v = None
      max_v = None
      with io.open(path, 'r', newline='', encoding=desc.get('encoding', 'utf-8')) as csvfile:
        reader = csv.reader(csvfile, delimiter=desc.get('separator', ','), quotechar=str(desc.get('quotechar', '|')))
        for row in reader:
          if cols is None:
            cols = len(row) - 1
          else:
            rows += 1
            min_act = min((float(f) for f in row[1:]))
            min_v = min_act if min_v is None else min(min_act, min_v)
            max_act = max((float(f) for f in row[1:]))
            max_v = max_act if max_v is None else max(max_act, max_v)
      desc['size'] = [rows, cols]
      desc['value']['range'] = [float(data['value_min']) if 'value_min' in data else min_v,
                                float(data['value_max']) if 'value_max' in data else max_v]

    return CSVMatrix(desc, project)


class CSVColumn(AColumn):
  def __init__(self, desc, table):
    super(CSVColumn, self).__init__(desc['name'], desc['value']['type'])
    self._desc = desc
    self._table = table

  def asnumpy(self, range=None):
    import pandas as pd
    p = self._table.aspandas(range)[self.name]
    if isinstance(p, pd.Series):
      return p.values
    return np.array([p])

  def process(self, index, data):
    is_number = self.type == 'real' or self.type == 'int'
    if is_number:
      return [np.NaN if d[index] == 'NA' or d[index] == '' else d[index] for d in data]
    else:
      return [d[index] for d in data]

  def dump(self):
    return self._desc


class CSVTable(CSVEntryMixin, ATable):
  def __init__(self, desc, project):
    ATable.__init__(self, desc['name'], project.id, desc['type'], desc.get('id', None))
    CSVEntryMixin.__init__(self, desc, project)
    desc['fqname'] = self.fqname
    desc['id'] = self.id
    self.idtype = desc['idtype']
    self.columns = [CSVColumn(d, self) for d in desc['columns']]
    self.shape = desc['size']

  def _process(self, data):
    rows = np.array([x[0] for x in data[1:]])
    import pandas as pd
    objs = {c.name: c.process(i + 1, data[1:]) for i, c in enumerate(self.columns)}
    df = pd.DataFrame(objs, columns=[c.name for c in self.columns])
    df.index = rows
    return {'rows': rows,
            'rowIds': assign_ids(rows, self.idtype),
            'df': df
            }

  def rows(self, range=None):
    n = self.load()['rows']
    if range is None:
      return n
    return n[range.asslice()]

  def rowids(self, range=None):
    n = self.load()['rowIds']
    if range is None:
      return n
    return n[range.asslice()]

  def aspandas(self, range=None):
    n = self.load()['df']
    if range is None:
      return n
    return n.iloc[range.asslice(no_ellipsis=True)]

  @staticmethod
  def parse(data, path, id=None):
    pass


class CSVVector(CSVEntryMixin, AVector):
  def __init__(self, desc, project):
    AVector.__init__(self, desc['name'], project.id, desc['type'], desc.get('id', None))
    CSVEntryMixin.__init__(self, desc, project)
    desc['fqname'] = self.fqname
    self.idtype = desc['idtype']
    self.value = desc['value']['type']
    self.range = desc['value']['range']
    self.shape = desc['size']

  def _process(self, data):
    is_number = self.value == 'real' or self.value == 'int'

    if is_number:
      data = [np.NaN if x[1] == 'NA' or x[1] == '' else x[1] for x in data[1:]]
    else:
      data = [x[1] for x in data[1:]]
    rows = np.array([x[0] for x in data[1:]])
    return {'rows': rows,
            'rowIds': assign_ids(rows, self.idtype),
            'data': np.array(data)
            }

  def rows(self, range=None):
    n = self.load()['rows']
    if range is None:
      return n
    return n[range.asslice()]

  def rowids(self, range=None):
    n = self.load()['rowIds']
    if range is None:
      return n
    return n[range.asslice()]

  def asnumpy(self, range=None):
    n = self.load()['data']
    if range is None:
      return n
    d = n[range[0].asslice()]
    if d.ndim == 0:
      d = d.reshape((1,))
    return d

  @staticmethod
  def parse(data, path, project, id=None):
    pass


def to_files(plugins):
  for plugin in plugins:
    index = os.path.join(plugin.folder + '/data/' if not hasattr(plugin, 'inplace') else plugin.folder, 'index.json')
    if not os.path.isfile(index):
      continue
    with open(index, 'r') as f:
      desc = json.load(f)
      for di in desc:
        if di['type'] == 'matrix':
          yield CSVMatrix(di, plugin)
        elif di['type'] == 'table':
          yield CSVTable(di, plugin)
        elif di['type'] == 'vector':
          yield CSVVector(di, plugin)
        elif di['type'] == 'stratification':
          yield CSVStratification(di, plugin)


class DataPlugin(object):
  def __init__(self, folder):
    # add a magic plugin for the static data dir
    self.inplace = True  # avoid adding the data suffix
    self.folder = folder
    self.id = os.path.basename(folder)

  def save(self, f):
    import werkzeug.utils
    from .util import random_id
    if not os.path.exists(self.folder):
      os.makedirs(self.folder)
    filename = os.path.basename(f.filename)
    filename = werkzeug.utils.secure_filename(filename + random_id(3) + '.csv')
    path = os.path.join(self.folder, filename)
    f.save(path)
    return path

  def append(self, desc, path):
    desc['path'] = os.path.basename(path)
    index = os.path.join(self.folder, 'index.json')
    old = []
    if os.path.isfile(index):
      with io.open(index, 'r', newline='', encoding=desc.get('encoding', 'utf-8')) as f:
        old = json.load(f)
    old.append(desc)
    with io.open(index, 'w', newline='', encoding=desc.get('encoding', 'utf-8')) as f:
      json.dump(old, f, indent=1)


class StaticFileProvider(ADataSetProvider):
  def __init__(self, plugins):

    self.files = list(to_files(plugins))

    cc = config.view('phovea_server')
    self.data_plugin = DataPlugin(os.path.join(cc.dataDir, 'data'))
    self.files.extend(to_files([self.data_plugin]))
    import glob
    extras = [DataPlugin(f) for f in (os.path.dirname(f) for f in glob.glob(cc.dataDir + '/*/index.json')) if
              os.path.basename(f) != 'data']
    self.files.extend(to_files(extras))

  def __iter__(self):
    return iter((f for f in self.files if f.can_read()))

  def upload(self, data, files, id=None):
    if 'csv' != data.get('_provider', 'csv'):
      return None  # not the right provider
    type = data.get('type', 'unknown')
    parsers = dict(matrix=CSVMatrix.parse, table=CSVTable.parse, vector=CSVVector.parse,
                   stratification=CSVStratification.parse)
    if type not in parsers:
      return None  # unknown type
    f = files[list(files.keys())[0]]
    path = self.data_plugin.save(f)
    r = parsers[type](data, path, self.data_plugin, id)
    if r:
      self.data_plugin.append(r._desc, path)
      self.files.append(r)
    else:
      os.remove(path)  # delete file again
    return r


def create():
  """
  entry point of this plugin
  """
  from .plugin import plugins
  return StaticFileProvider(plugins())



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import zip
from builtins import str
from . import ns
from . import range as ranges
from .util import jsonify
from .plugin import list as list_plugins
import numpy as np
from PIL import Image


def asrange(r):
  if r is None or r == '':
    return None
  return ranges.parse(r)


def format_json(dataset, range, args):
  d = dataset.asjson(range)
  if bool(args.get('f_pretty_print', False)):
    return jsonify(d, indent=' ', allow_nan=False)
  return jsonify(d, allow_nan=False)


def format_csv(dataset, range, args):  # noqa
  include_rows = bool(args.get('f_rows', False))
  include_cols = bool(args.get('f_cols', False))
  delimiter = args.get('f_delimiter', ';')

  import numpy as np
  import numpy.ma as ma

  def to_str(v):
    if isinstance(v, str):
      return v
    if np.isnan(v) or ma.is_masked(v):
      return ''
    return str(v)

  def gen():
    if include_cols and dataset.type == 'matrix':
      cols = dataset.cols(range[1] if range is not None else None)
      header = delimiter.join(cols)
    elif dataset.type == 'table':
      header = delimiter.join([d.name for d in dataset.columns])
    else:
      header = ''

    d = dataset.aslist(range)

    if include_rows:
      rows = dataset.rows(range[0] if range is not None else None)
      yield dataset.idtype if dataset.type == 'table' else dataset.rowtype
      yield delimiter

    yield header
    yield '\n'

    if include_rows:
      # extend with the row ids
      for row, line in zip(rows, d):
        yield row
        yield delimiter
        l = [to_str(v) for v in line] if dataset.type == 'matrix' else (to_str(line[d.name] for d in dataset.columns))
        yield delimiter.join(l)
        yield '\n'
    else:
      for line in d:
        l = [to_str(v) for v in line] if dataset.type == 'matrix' else (to_str(line[d.name] for d in dataset.columns))
        yield delimiter.join(l)
        yield '\n'

  return ns.Response(gen(), mimetype='text/csv',
                     headers={'Content-Disposition': 'attachment;filename=' + dataset.name + '.csv'})


def _hex_to_rgb(hex):
  hex = hex.lstrip('#')
  # hlen = len(hex)
  hlen = 6  # assume 6-digits HEX value! Does not work with 3- or 8-digits HEX values!
  return tuple(int(hex[i:i + hlen // 3], 16) for i in range(0, hlen, hlen // 3))


def _color_palette(arg):
  if arg is None or arg == '':
    return None
  if arg == 'blue_white_red':
    from .colors import blue_white_red
    return blue_white_red.as_palette()
  elif arg == 'white_red':
    from .colors import white_red
    return white_red.as_palette()

  # generate color palette
  from .colors import ColorPalette
  colors = arg.split('-')
  colors = [_hex_to_rgb(c) for c in colors]
  return ColorPalette(*colors).as_palette()


def _set_missing_values(img, arr, color):
  import numpy as np
  locs = np.transpose(np.where(np.isnan(arr)))
  if locs.size > 0:
    img = img.convert('RGB')
    for loc in locs:
      img.putpixel((loc[1], loc[0]), color)
  return img

# bytescale and scipy.misc.toimage() were completely removed -> the original functions are ported here

_errstr = "Mode is unknown or incompatible with input array shape."


def bytescale(data, cmin=None, cmax=None, high=255, low=0):
    """
    Byte scales an array (image).
    Byte scaling means converting the input image to uint8 dtype and scaling
    the range to ``(low, high)`` (default 0-255).
    If the input image already has dtype uint8, no scaling is done.
    This function is only available if Python Imaging Library (PIL) is installed.
    Parameters
    ----------
    data : ndarray
        PIL image data array.
    cmin : scalar, optional
        Bias scaling of small values. Default is ``data.min()``.
    cmax : scalar, optional
        Bias scaling of large values. Default is ``data.max()``.
    high : scalar, optional
        Scale max value to `high`.  Default is 255.
    low : scalar, optional
        Scale min value to `low`.  Default is 0.
    Returns
    -------
    img_array : uint8 ndarray
        The byte-scaled array.
    """
    if data.dtype == np.uint8:
        return data

    if high > 255:
        raise ValueError("`high` should be less than or equal to 255.")
    if low < 0:
        raise ValueError("`low` should be greater than or equal to 0.")
    if high < low:
        raise ValueError("`high` should be greater than or equal to `low`.")

    if cmin is None:
        cmin = data.min()
    if cmax is None:
        cmax = data.max()

    cscale = cmax - cmin
    if cscale < 0:
        raise ValueError("`cmax` should be larger than `cmin`.")
    elif cscale == 0:
        cscale = 1

    scale = float(high - low) / cscale
    bytedata = (data - cmin) * scale + low
    return (bytedata.clip(low, high) + 0.5).astype(np.uint8)


def toimage(arr, high=255, low=0, cmin=None, cmax=None, pal=None,
            mode=None, channel_axis=None):
    """Takes a numpy array and returns a PIL image.
    This function is only available if Python Imaging Library (PIL) is installed.
    The mode of the PIL image depends on the array shape and the `pal` and
    `mode` keywords.
    For 2-D arrays, if `pal` is a valid (N,3) byte-array giving the RGB values
    (from 0 to 255) then ``mode='P'``, otherwise ``mode='L'``, unless mode
    is given as 'F' or 'I' in which case a float and/or integer array is made.
    .. warning::
        This function uses `bytescale` under the hood to rescale images to use
        the full (0, 255) range if ``mode`` is one of ``None, 'L', 'P', 'l'``.
        It will also cast data for 2-D images to ``uint32`` for ``mode=None``
        (which is the default).
    Notes
    -----
    For 3-D arrays, the `channel_axis` argument tells which dimension of the
    array holds the channel data.
    For 3-D arrays if one of the dimensions is 3, the mode is 'RGB'
    by default or 'YCbCr' if selected.
    The numpy array must be either 2 dimensional or 3 dimensional.
    """
    data = np.asarray(arr)
    if np.iscomplexobj(data):
        raise ValueError("Cannot convert a complex-valued array.")
    shape = list(data.shape)
    valid = len(shape) == 2 or ((len(shape) == 3) and ((3 in shape) or (4 in shape)))
    if not valid:
        raise ValueError("'arr' does not have a suitable array shape for "
                         "any mode.")
    if len(shape) == 2:
        shape = (shape[1], shape[0])  # columns show up first
        if mode == 'F':
            data32 = data.astype(np.float32)
            image = Image.frombytes(mode, shape, data32.tostring())
            return image
        if mode in [None, 'L', 'P']:
            bytedata = bytescale(data, high=high, low=low,
                                 cmin=cmin, cmax=cmax)
            image = Image.frombytes('L', shape, bytedata.tostring())
            if pal is not None:
                image.putpalette(np.asarray(pal, dtype=np.uint8).tostring())
                # Becomes a mode='P' automatically.
            elif mode == 'P':  # default gray-scale
                pal = (np.arange(0, 256, 1, dtype=np.uint8)[:, np.newaxis] * np.ones((3,), dtype=np.uint8)[np.newaxis, :])
                image.putpalette(np.asarray(pal, dtype=np.uint8).tostring())
            return image
        if mode == '1':  # high input gives threshold for 1
            bytedata = (data > high)
            image = Image.frombytes('1', shape, bytedata.tostring())
            return image
        if cmin is None:
            cmin = np.amin(np.ravel(data))
        if cmax is None:
            cmax = np.amax(np.ravel(data))
        data = (data*1.0 - cmin)*(high - low)/(cmax - cmin) + low
        if mode == 'I':
            data32 = data.astype(np.uint32)
            image = Image.frombytes(mode, shape, data32.tostring())
        else:
            raise ValueError(_errstr)
        return image

    # if here then 3-d array with a 3 or a 4 in the shape length.
    # Check for 3 in datacube shape --- 'RGB' or 'YCbCr'
    if channel_axis is None:
        if (3 in shape):
            ca = np.flatnonzero(np.asarray(shape) == 3)[0]
        else:
            ca = np.flatnonzero(np.asarray(shape) == 4)
            if len(ca):
                ca = ca[0]
            else:
                raise ValueError("Could not find channel dimension.")
    else:
        ca = channel_axis

    numch = shape[ca]
    if numch not in [3, 4]:
        raise ValueError("Channel axis dimension is not valid.")

    bytedata = bytescale(data, high=high, low=low, cmin=cmin, cmax=cmax)
    if ca == 2:
        strdata = bytedata.tostring()
        shape = (shape[1], shape[0])
    elif ca == 1:
        strdata = np.transpose(bytedata, (0, 2, 1)).tostring()
        shape = (shape[2], shape[0])
    elif ca == 0:
        strdata = np.transpose(bytedata, (1, 2, 0)).tostring()
        shape = (shape[2], shape[1])
    if mode is None:
        if numch == 3:
            mode = 'RGB'
        else:
            mode = 'RGBA'

    if mode not in ['RGB', 'RGBA', 'YCbCr', 'CMYK']:
        raise ValueError(_errstr)

    if mode in ['RGB', 'YCbCr']:
        if numch != 3:
            raise ValueError("Invalid array shape for mode.")
    if mode in ['RGBA', 'CMYK']:
        if numch != 4:
            raise ValueError("Invalid array shape for mode.")

    # Here we know data and mode is correct
    image = Image.frombytes(mode, shape, strdata)
    return image

# end of ported functions


def format_image(dataset, range, args):
  format = args.get('format', 'png')

  import io
  import numpy as np

  # TODO set a palette to specify colors instead of gray scales
  # how to interpolate / sample colors - which space?
  minmax = dataset.range
  cmin = float(args.get('format_min', minmax[0]))
  cmax = float(args.get('format_max', minmax[1]))
  d = np.array(dataset.aslist(range))
  if d.ndim == 1:
    d = d.reshape((1, d.shape[0]))
  img = toimage(d, cmin=cmin, cmax=cmax, pal=_color_palette(args.get('format_palette', None)))

  # convert to real RGB image
  # inject missing values
  img = _set_missing_values(img, d, _hex_to_rgb(args.get('format_missing'[0], '#d400c2')))

  if 'format_w' in args:
    width = int(args.get('format_w'))
    wpercent = width // float(img.size[0])
    height = int(args.get('format_h', (float(img.size[1]) * float(wpercent))))
    from PIL.Image import NEAREST
    img = img.resize((width, height), NEAREST)
  elif 'format_h' in args:
    height = int(args.get('format_h'))
    hpercent = height // float(img.size[1])
    width = int(float(img.size[0]) * float(hpercent))
    from PIL.Image import NEAREST
    img = img.resize((width, height), NEAREST)

  if args.get('format_transpose', False):
    from PIL.Image import ROTATE_90
    img = img.transpose(ROTATE_90)

  b = io.BytesIO()
  img.save(b, format=format)
  b.seek(0)
  return ns.send_file(b, mimetype='image/' + format.replace('jpg', 'jpeg'))


def resolve_formatter(type, format):
  for p in list_plugins(type + '-formatter'):
    if p.format == format:
      return p.load()
  desc = ','.join((p.format for p in list_plugins(type + '-formatter')))
  ns.abort(400, 'unknown format "{0}" possible formats are: {1}'.format(format, desc))


def _add_handler(app, dataset_getter, type):
  def desc_gen(dataset_id):
    d = dataset_getter(dataset_id, type)
    return jsonify(d.to_description())

  app.add_url_rule('/' + type + '/<dataset_id>', 'desc_' + type, ns.etag(desc_gen))

  def rows_gen(dataset_id):
    d = dataset_getter(dataset_id, type)
    r = asrange(ns.request.args.get('range', None))
    return jsonify(d.rows(r[0] if r is not None else None))

  app.add_url_rule('/' + type + '/<dataset_id>/rows', 'rows_' + type, ns.etag(rows_gen))

  def rowids_gen(dataset_id):
    d = dataset_getter(dataset_id, type)
    r = asrange(ns.request.args.get('range', None))
    ids = d.rowids(r[0] if r is not None else None)
    return jsonify(str(ranges.from_list(list(ids))))

  app.add_url_rule('/' + type + '/<dataset_id>/rowIds', 'rowids_' + type, ns.etag(rowids_gen))

  def raw_gen(dataset_id):
    d = dataset_getter(dataset_id, type)
    r = asrange(ns.request.args.get('range', None))
    return jsonify(d.aslist(r), allow_nan=False)

  app.add_url_rule('/' + type + '/<dataset_id>/raw', 'raw_' + type, ns.etag(raw_gen))

  def data_gen(dataset_id):
    d = dataset_getter(dataset_id, type)
    r = asrange(ns.request.values.get('range', None))
    formatter = resolve_formatter(type, ns.request.values.get('format', 'json'))
    return formatter(d, r, args=ns.request.values)

  app.add_url_rule('/' + type + '/<dataset_id>/data', 'data_' + type, ns.etag(data_gen), methods=['GET', 'POST'])


def add_table_handler(app, dataset_getter):
  _add_handler(app, dataset_getter, 'table')

  def find_view(dataset_id, view_name):
    d = dataset_getter(dataset_id, 'table')
    if hasattr(d, 'views') and view_name in d.views:
      view = d.views[view_name]
      args = ns.request.args.to_dict()
      return view, args
    ns.abort(404)

  def col_table(dataset_id, column):
    d = dataset_getter(dataset_id, 'table')
    r = asrange(ns.request.args.get('range', None))
    for col in d.columns:
      if col.name == column or col.dump().get('column', '') == column:
        return jsonify(col.aslist(r), allow_nan=False)
    ns.abort(404)

  def view_table(dataset_id, view_name):
    view, args = find_view(dataset_id, view_name)
    formatter = resolve_formatter('table', ns.request.args.get('format', 'json'))
    return formatter(view, args, args=ns.request.args)

  def view_raw_table(dataset_id, view_name):
    view, args = find_view(dataset_id, view_name)
    return jsonify(view.aslist(args), allow_nan=False)

  def view_rows_table(dataset_id, view_name):
    view, args = find_view(dataset_id, view_name)
    return jsonify(view.rows(args))

  def view_rowids_table(dataset_id, view_name):
    view, args = find_view(dataset_id, view_name)
    ids = view.rowids(args)
    return jsonify(str(ranges.from_list(list(ids))))

  app.add_url_rule('/table/<dataset_id>/col/<column>', 'col_table', ns.etag(col_table))

  app.add_url_rule('/table/<dataset_id>/view/<view_name>', 'view_table', ns.etag(view_table))
  app.add_url_rule('/table/<dataset_id>/view/<view_name>/raw', 'view_raw_table', ns.etag(view_raw_table))
  app.add_url_rule('/table/<dataset_id>/view/<view_name>/rows', 'view_rows_table', ns.etag(view_rows_table))
  app.add_url_rule('/table/<dataset_id>/view/<view_name>/rowIds', 'view_rowids_table', ns.etag(view_rowids_table))


def _stats_of(data):
  import numpy as np
  import scipy.stats
  return dict(median=np.nanmedian(data),
              q1=np.nanpercentile(data, 25),
              q3=np.nanpercentile(data, 75),
              min=np.nanmin(data),
              max=np.nanmax(data),
              sum=np.nansum(data),
              mean=np.nanmean(data),
              var=np.nanvar(data),
              sd=np.nanstd(data),
              n=len(data),
              nans=np.count_nonzero(np.isnan(data)),
              moment2=scipy.stats.moment(data, 2),
              moment3=scipy.stats.moment(data, 3),
              moment4=scipy.stats.moment(data, 4),
              kurtosis=scipy.stats.kurtosis(data),
              skewness=scipy.stats.skew(data))


def add_vector_handler(app, dataset_getter):
  _add_handler(app, dataset_getter, 'vector')

  def hist_vector(dataset_id):
    d = dataset_getter(dataset_id, 'vector')
    r = asrange(ns.request.args.get('range', None))
    import numpy as np
    data = d.asnumpy(r)
    hist, bin_edges = np.histogram(data, bins=int(ns.request.args.get('bins', np.sqrt(len(data)))), range=d.range)
    return jsonify(hist)

  def stats_vector(dataset_id):
    d = dataset_getter(dataset_id, 'vector')
    r = asrange(ns.request.args.get('range', None))
    data = d.asnumpy(r)
    return jsonify(_stats_of(data))

  app.add_url_rule('/vector/<dataset_id>/hist', 'hist_vector', ns.etag(hist_vector))
  app.add_url_rule('/vector/<dataset_id>/stats', 'stats_vector', ns.etag(stats_vector))


def add_matrix_handler(app, dataset_getter):
  """
  add the handlers for handling a matrix
  :param app:
  :param dataset_getter:
  :return:
  """
  _add_handler(app, dataset_getter, 'matrix')

  def cols_matrix(dataset_id):
    d = dataset_getter(dataset_id, 'matrix')
    r = asrange(ns.request.args.get('range', None))
    return jsonify(d.cols(r[0] if r is not None else None))

  app.add_url_rule('/matrix/<dataset_id>/cols', 'cols_matrix', ns.etag(cols_matrix))

  def colids_matrix(dataset_id):
    d = dataset_getter(dataset_id, 'matrix')
    r = asrange(ns.request.args.get('range', None))
    ids = d.colids(r[0] if r is not None else None)
    return jsonify(str(ranges.from_list(list(ids))))

  app.add_url_rule('/matrix/<dataset_id>/colIds', 'colids_matrix', ns.etag(colids_matrix))

  def hist_matrix(dataset_id):
    d = dataset_getter(dataset_id, 'matrix')
    r = asrange(ns.request.args.get('range', None))
    data = d.asnumpy(r)
    import numpy as np
    hist, bin_edges = np.histogram(data, bins=int(ns.request.args.get('bins', np.sqrt(len(data)))), range=d.range)
    return jsonify(hist)

  def stats_matrix(dataset_id):
    d = dataset_getter(dataset_id, 'matrix')
    r = asrange(ns.request.args.get('range', None))
    data = d.asnumpy(r)
    return jsonify(_stats_of(data.flat))

  app.add_url_rule('/matrix/<dataset_id>/hist', 'hist_matrix', ns.etag(hist_matrix))
  app.add_url_rule('/matrix/<dataset_id>/stats', 'stats_matrix', ns.etag(stats_matrix))



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object
from flask_sockets import Sockets as Socket  # noqa 
from geventwebsocket import WebSocketError  # noqa 
import logging

_log = logging.getLogger(__name__)


class WebsocketSend(object):
  def __init__(self, ws):
    self._ws = ws

  def send(self, msg_type, payload):
    import pandas.json as ujson
    d = ujson.dumps(dict(type=msg_type, data=payload))
    self._ws.send(d)

  def send_str(self, msg_type, payload_as_string):
    d = '{{ "type": "{}", "data": {}}}'.format(msg_type, payload_as_string)
    self._ws.send(d)


def websocket_loop(ws, handler_map):
  import pandas.json as ujson
  while True:
    msg = ws.receive()
    if msg is None:
      continue
    _log.debug('msg received %s', msg)
    data = ujson.loads(msg)
    t = data['type']
    payload = data['data']

    if t not in handler_map:
      _log.warning('no handler defined for message of type: %s', t)
    else:
      handler_map[t](payload, WebsocketSend(ws))



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import map
from builtins import str
from builtins import range
from builtins import object
from itertools import islice
import logging

_log = logging.getLogger(__name__)


class MemoryIDAssigner(object):
  """
  assigns ids to object in memory only, i.e. not persisted
  """

  def __init__(self):
    self._idsmapping = {}

  def unmap(self, uids, idtype):
    if idtype not in self._idsmapping:
      return [None] * len(uids)
    cache = self._idsmapping[idtype]

    def lookup(id):
      for k, v in cache.items():
        if v == id:
          return k
      return None

    return list(map(lookup, uids))

  def load(self, idtype, mapping):
    """
    resets and loads the given mapping
    :param idtype:
    :param mapping: array of tuples (id, uid)
    :return:
    """
    del self._idsmapping[idtype]
    # assuming incremental ids
    self._idsmapping[idtype] = {id: uid for id, uid in mapping}

  def search(self, idtype, query, max_results=None):
    if idtype not in self._idsmapping:
      return []
    mappings = self._idsmapping[idtype]
    query = query.lower()

    return list(islice((dict(id=v, name=k) for k, v in mappings.items() if query in k.lower()), max_results))

  def __call__(self, ids, idtype):
    """
     return the integer index ids for the given ids in the given idtype
    """
    if idtype not in self._idsmapping:
      self._idsmapping[idtype] = {id: i for i, id in enumerate(ids)}
      # 1 to 1 mapping
      return list(range(len(ids)))

    cache = self._idsmapping[idtype]

    def add(id):
      i = cache.get(id, -1)
      if i < 0:  # not yet part of
        i = len(cache)
        cache[id] = i
      return i

    return [add(id) for id in ids]


class DBIDAssigner(object):
  """
  assigns ids to object using a dbm database
  """

  def __init__(self):
    import dbm
    import phovea_server.config
    import os
    base_dir = phovea_server.config.get('phovea_server.dataDir')
    if not os.path.exists(base_dir):
      os.makedirs(base_dir)
    self._db = dbm.open(base_dir + '/mapping.dbm', 'c')

  @staticmethod
  def to_forward_key(idtype, identifier):
    return idtype + '2id.' + str(identifier)

  @staticmethod
  def to_backward_key(idtype, id):
    return 'id2' + idtype + '.' + str(id)

  def unmap(self, uids, idtype):

    def lookup(id):
      key = self.to_backward_key(idtype, id)
      return self._db.get(key, None)

    return list(map(lookup, uids))

  def load(self, idtype, mapping):
    """
    resets and loads the given mapping
    :param idtype:
    :param mapping: array of tuples (id, uid)
    :return:
    """
    # assuming incremental ids
    if idtype in self._db:
      # clear old data
      for key in list(self._db.keys()):
        if key.startswith(idtype + '2id.') or key.startswith('id2' + idtype + '.'):
          del self._db[key]

    max_uid = None
    for id, uid in mapping:
      key = self.to_forward_key(idtype, id)
      max_uid = uid if max_uid is None else max(uid, max_uid)
      self._db[key] = str(uid)
      self._db[self.to_backward_key(idtype, uid)] = str(id)

  def search(self, idtype, query, max_results=None):
    from fnmatch import fnmatch
    pattern = idtype + '2id.*' + query + '*'

    return list(islice((dict(id=v, name=k) for k, v in self._db.items() if fnmatch(k, pattern)), max_results))

  def __call__(self, ids, idtype):
    """
     return the integer index ids for the given ids in the given idtype
    """
    max_old = -1 if idtype not in self._db else int(self._db[idtype])
    r = []
    for id in ids:
      key = self.to_forward_key(idtype, id)
      if key in self._db:
        r.append(int(self._db[key]))
      else:
        i = max_old + 1
        _log.debug('create %s %d', key, i)
        max_old += 1
        self._db[key] = str(i)
        self._db[self.to_backward_key(idtype, i)] = str(id)
        r.append(i)

    self._db[idtype] = str(max_old)
    return r


class SqliteIDAssigner(object):
  """
  assigns ids to object using a sqlite database
  """

  def __init__(self):
    import sqlite3
    import phovea_server.config
    import os
    base_dir = phovea_server.config.get('phovea_server.dataDir')
    if not os.path.exists(base_dir):
      os.makedirs(base_dir)
    self._db = sqlite3.connect(base_dir + '/mapping.sqlite3')
    self._db.execute('CREATE TABLE IF NOT EXISTS mapping(idtype TEXT, name TEXT, id INT, PRIMARY KEY(idtype,name), UNIQUE (idtype,id))')
    self._cache = {}

  def unmap(self, uids, idtype):
    existing = self.get_cache(idtype)

    def lookup(id):
      for k, v in existing.items():
        if v == id:
          return k
      return None

    return list(map(lookup, uids))

  def get_cache(self, idtype):
    existing = self._cache.get(idtype, None)

    # lazy load
    if existing is None:
      existing = dict()
      self._cache[idtype] = existing
      for row in self._db.execute('SELECT name, id FROM mapping WHERE idtype=?', (idtype,)):
        existing[row[0]] = row[1]

    return existing

  def load(self, idtype, mapping):
    """
    resets and loads the given mapping
    :param idtype:
    :param mapping: array of tuples (id, uid)
    :return:
    """
    # delete cache
    del self._cache[idtype]
    self._db.execute('DELETE FROM mapping WHERE idtype=?', (idtype,))

    self._db.executemany('insert or ignore into mapping values ("' + idtype + '",?,?)', mapping)
    self._db.commit()
    self._cache[idtype] = {id: uid for id, uid in mapping}

  def __call__(self, ids, idtype):
    """
     return the integer index ids for the given ids in the given idtype
    """
    existing = self.get_cache(idtype)

    missing = []
    r = []
    max_v = len(existing) - 1

    for id in ids:
      id = str(id)
      if id in existing:
        i = existing[id]
      else:
        max_v += 1
        i = max_v
        missing.append((id, i))
        existing[id] = i
      r.append(i)

    if len(missing) > 0:
      _log.debug('add missing to %s %d', idtype, len(missing))
      self._db.executemany('insert or ignore into mapping values ("' + idtype + '",?,?)', missing)
      self._db.commit()

    return r

  def search(self, idtype, query, max_results=None):
    mappings = self.get_cache(idtype)
    query = query.lower()
    return list(islice((dict(id=v, name=k) for k, v in mappings.items() if query in k.lower()), max_results))


# TODO implement a real id mappper

_assigner = SqliteIDAssigner()


def create():
  return _assigner



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################

from builtins import object
import abc
import numpy as np


def to_plural(s):
  if s[len(s) - 1] == 'y':
    return s[0:len(s) - 1] + 'ies'
  return s + 's'


def to_idtype_description(id):
  return dict(id=id, name=id, names=to_plural(id))


class ADataSetEntry(object, metaclass=abc.ABCMeta):
  """
  A basic dataset entry
  """

  def __init__(self, name, project, type, id=None):
    """
    constructor for a new dataset
    :param name:
    :param project: the parent/folder of this dataset
    :param type: the type of this dataset
    :param id: optional the id to use
    """
    self.name = name
    self.fqname = project + '/' + name
    self.type = type
    from .util import fix_id
    self.id = id if id is not None else fix_id(self.fqname)

  def idtypes(self):
    """
    :return: the list of all idtypes as string array
    """
    return []

  def to_description(self):
    """
    :return: a dictionary describing this dataset in a client understandable format
    """
    return dict(type=self.type,
                name=self.name,
                id=self.id,
                fqname=self.fqname)

  def to_idtype_descriptions(self):
    """
    list of a all idtypes of this dataset
    :return:
    """

    return [to_idtype_description(t) for t in self.idtypes()]

  def update(self, args, files):
    """
    updates this dataset with the new data
    :param args: data dict
    :param files: list of FileStorage files
    :return: boolean whether the operation was successful
    """
    return False

  def modify(self, args, files):
    """
    modifies this dataset with the given arguments
    :param args: data dict
    :param files: list of FileStorage files
    :return: boolean whether the operation was successful
    """
    return False

  def remove(self):
    """
    removes itself
    :return: boolean whether it was successfully removed
    """
    return False

  @abc.abstractmethod
  def asjson(self, range=None):
    """
    converts this dataset to a json compatible format
    :param range: optional sub range to deliver
    :return: a json compatible dataset representation
    """
    return dict()

  def can_read(self, user=None):
    from .security import can_read
    return can_read(self.to_description(), user)

  def can_write(self, user=None):
    from .security import can_write
    return can_write(self.to_description(), user)


class AStratification(ADataSetEntry, metaclass=abc.ABCMeta):
  """
  A basic dataset entry
  """

  def __init__(self, name, project, type, id=None):
    super(AStratification, self).__init__(name, project, type, id)
    self.idtype = 'Custom'

  @abc.abstractmethod
  def rows(self, range=None):
    return []

  @abc.abstractmethod
  def rowids(self, range=None):
    return []

  @abc.abstractmethod
  def groups(self):
    return []

  def idtypes(self):
    return [self.idtype]


class AMatrix(ADataSetEntry, metaclass=abc.ABCMeta):
  """
  A basic dataset entry
  """

  def __init__(self, name, project, type, id=None):
    super(AMatrix, self).__init__(name, project, type, id)
    self.rowtype = 'Custom'
    self.coltype = 'Custom'
    self.shape = [0, 0]
    self.value = 'string'

  @abc.abstractmethod
  def rows(self, range=None):
    return []

  @abc.abstractmethod
  def rowids(self, range=None):
    return []

  @abc.abstractmethod
  def cols(self, range=None):
    return []

  @abc.abstractmethod
  def colids(self, range=None):
    return []

  def aslist(self, range=None):
    return self.asnumpy(range).tolist()

  @abc.abstractmethod
  def asnumpy(self, range=None):
    return np.array([])

  def asjson(self, range=None):
    arr = self.asnumpy(range)
    rows = self.rows(None if range is None else range[0])
    cols = self.cols(None if range is None else range[1])
    rowids = self.rowids(None if range is None else range[0])
    colids = self.colids(None if range is None else range[1])

    r = dict(data=arr, rows=rows, cols=cols, rowIds=rowids, colIds=colids)
    return r

  def idtypes(self):
    return [self.rowtype, self.coltype]


class AVector(ADataSetEntry, metaclass=abc.ABCMeta):
  """
  A basic dataset entry
  """

  def __init__(self, name, project, type, id=None):
    super(AVector, self).__init__(name, project, type, id)
    self.idtype = 'Custom'
    self.value = 'string'
    self.shape = [0]

  @abc.abstractmethod
  def rows(self, range=None):
    return []

  @abc.abstractmethod
  def rowids(self, range=None):
    return []

  def aslist(self, range=None):
    return self.asnumpy(range).tolist()

  @abc.abstractmethod
  def asnumpy(self, range=None):
    return np.array([])

  def asjson(self, range=None):
    arr = self.asnumpy(range)
    rows = self.rows(None if range is None else range[0])
    rowids = self.rowids(None if range is None else range[0])
    r = dict(data=arr, rows=rows, rowIds=rowids)

    return r

  def idtypes(self):
    return [self.idtype]


class AColumn(object, metaclass=abc.ABCMeta):
  def __init__(self, name, type):
    self.name = name
    self.type = type

  def aslist(self, range=None):
    return self.asnumpy(range).tolist()

  @abc.abstractmethod
  def asnumpy(self, range=None):
    return np.array([])

  @abc.abstractmethod
  def dump(self):
    return None


class ATable(ADataSetEntry, metaclass=abc.ABCMeta):
  """
  A basic dataset entry
  """

  def __init__(self, name, project, type, id=None):
    super(ATable, self).__init__(name, project, type, id)
    self.idtype = 'Custom'
    self.shape = [0, 0]
    self.columns = []

  @abc.abstractmethod
  def rows(self, range=None):
    return []

  @abc.abstractmethod
  def rowids(self, range=None):
    return []

  def aslist(self, range=None):
    data = self.aspandas(range)

    # if range is only one element, aspandas returns a Series, otherwise a Dataframe
    # to_dict('records') throws error on Series
    if range is not None and range.dims[0].size() == 1:
      return data.to_dict()

    return data.to_dict('records')

  @abc.abstractmethod
  def aspandas(self, range=None):
    import pandas as pd
    return pd.DataFrame()

  def asjson(self, range=None):
    arr = self.aslist(range)
    rows = self.rows(None if range is None else range[0])
    rowids = self.rowids(None if range is None else range[0])
    r = dict(data=arr, rows=rows, rowIds=rowids)

    return r

  def idtypes(self):
    return [self.idtype]


class ADataSetProvider(object, metaclass=abc.ABCMeta):
  def __len__(self):
    import itertools
    return itertools.count(self)

  @abc.abstractmethod
  def __iter__(self):
    return iter([])

  def __getitem__(self, dataset_id):
    """
    get a specific dataset item by id
    :param dataset_id:
    :return: the dataset or None
    """
    for elem in self:
      if elem.id == dataset_id:
        return elem
    return None

  def remove(self, entry):
    return False

  def upload(self, data, files, id=None):
    """
    adds a new dataset to this provider
    :param data: the description data dict object
    :param files: a list of FileStorage files containing data files
    :param id: optional unique id of the newly created dataset
    :return: None if the element can't be uploaded else the dataset
    """
    return None



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import object
import logging
from functools import cmp_to_key

_registry = None


def _get_registry():
  global _registry
  if _registry is None:
    from ._plugin_parser import parse
    metadata = parse()
    from .config import merge_plugin_configs, _c, _initialize
    # check initialization
    if _c is None:
      _initialize()
    merge_plugin_configs(metadata.plugins)
    _registry = Registry(metadata.plugins, metadata.server_extensions, metadata)
  return _registry


class Extension(object):
  """
   the loaded plugin instance
  """

  def __init__(self, desc, impl):
    self.desc = desc
    self.impl = impl
    self._cache = None

  def __call__(self, *args, **kwargs):
    """
     access and call the factory method of this plugin
    """
    if getattr(self.desc, 'singleton', False) and self._cache is not None:
      return self._cache

    m = getattr(self.impl, self.desc.factory)

    if hasattr(m, '__call__'):
      v = m(*args, **kwargs)
    else:
      v = m
    self._cache = v
    return v

  def factory(self, *args, **kwargs):
    return self(*args, **kwargs)


class AExtensionDesc(object):
  def __init__(self, desc):
    self.type = desc.get('type', 'unknown')
    self.id = desc['id']
    self.name = self.id
    self.factory = 'create'
    self.file = 'main'
    self.version = '1.0'
    self.description = ''
    # copy all values
    for key, value in desc.items():
      self.__dict__[key] = value


class ExtensionDesc(AExtensionDesc):
  """
   plugin description
  """

  def __init__(self, desc):
    super(ExtensionDesc, self).__init__(desc)
    self._impl = None

    # from js notation to python notation
    self.module = self.module.replace('/', '.')

  def load(self):
    if self._impl is None:
      import importlib
      _log = logging.getLogger(__name__)
      _log.info('importing %s', self.module)

      m = importlib.import_module(self.module)
      if hasattr(m, '_plugin_initialize'):  # init method
        # import inspect
        # inspect.getcallargs()
        m._plugin_initialize()

      self._impl = Extension(self, m)
    return self._impl


class PreLoadedExtensionDesc(AExtensionDesc):
  def __init__(self, desc, impl):
    super(PreLoadedExtensionDesc, self).__init__(desc)
    self._wrapper = PreLoadedExtension(impl)

  def load(self):
    return self._wrapper


class PreLoadedExtension(object):
  def __init__(self, impl):
    self._impl = impl

  def __call__(self, *args, **kwargs):
    return self._impl

  def factory(self, *args, **kwargs):
    return self._impl


class Registry(object):
  def __init__(self, plugins, extensions, metadata):
    self.plugins = plugins
    self.metadata = metadata
    self._extensions = [ExtensionDesc(p) for p in extensions]
    self._extensions.append(PreLoadedExtensionDesc(dict(type='manager', id='registry'), self))

    self._singletons = None

  @property
  def singletons(self):
    import collections
    from . import config
    # check initialization
    _log = logging.getLogger(__name__)
    if self._singletons is not None:
      return self._singletons

    def loader(e):
      return lambda: e.load().factory()

    # select singleton impl with lowest priority default 100
    mm = collections.defaultdict(lambda: [])
    for e in self._extensions:
      if e.type == 'manager':
        mm[e.id].append(e)

    if config._c is None:
      config._initialize()
    cc = config.view('phovea_server._runtime')
    current_command = cc.get('command', default='unknown')
    _log.info('read currently executed command from config: %s', current_command)

    def compare(a, b):
      a_prio = getattr(a, 'priority', 100)
      a_command = getattr(a, 'command', None)
      b_prio = getattr(b, 'priority', 100)
      b_command = getattr(b, 'command', None)
      # if the command matches the current command this has priority
      if a_command != b_command:
        if a_command == current_command:
          return -1
        elif b_command == current_command:
          return 1
      return a_prio - b_prio

    def select(v):
      v = sorted(v, key=cmp_to_key(compare))
      _log.info('creating singleton %s %s', v[0].id, getattr(v[0], 'module', 'server'))
      return loader(v[0])

    self._singletons = {k: select(v) for k, v in mm.items()}

    return self._singletons

  def __len__(self):
    return len(self._extensions)

  def __getitem__(self, item):
    return self._extensions[item]

  def __iter__(self):
    return iter(self._extensions)

  def list(self, plugin_type=None):
    if plugin_type is None:
      return self
    if not hasattr(plugin_type, '__call__'):  # not a callable
      return [x for x in self if x.type == plugin_type]
    return [x for x in self if plugin_type(x)]

  def lookup(self, singleton_id):
    if singleton_id in self.singletons:
      return self.singletons[singleton_id]()
    return None


def list(plugin_type=None):
  return _get_registry().list(plugin_type)


def lookup(singleton_id):
  return _get_registry().lookup(singleton_id)


def plugins():
  return _get_registry().plugins


def metadata():
  return _get_registry().metadata



###############################################################################
# Caleydo - Visualization for Molecular Biology - http://caleydo.org
# Copyright (c) The Caleydo Team. All rights reserved.
# Licensed under the new BSD license, available at http://caleydo.org/license
###############################################################################


from builtins import str
import phovea_server.plugin
import phovea_server.range
import phovea_server.util
from phovea_server.dataset_def import to_idtype_description
import itertools

_providers_r = None


def _providers():
  global _providers_r
  if _providers_r is None:
    _providers_r = [p.load().factory() for p in phovea_server.plugin.list('dataset-provider')]
  return _providers_r


def iter():
  """
  an iterator of all known datasets
  :return:
  """
  return itertools.chain(*_providers())


def list_datasets():
  """
  list all known datasets
  :return:
  """
  return list(iter())


def get(dataset_id):
  """
  :param dataset_id:
  :return: returns the selected dataset identified by id
  """
  for p in _providers():
    r = p[dataset_id]
    if r is not None:
      return r
  return None


def add(desc, files=[], id=None):
  """
  adds a new dataset to this storage
  :param desc: the dict description information
  :param files: a list of FileStorage
  :param id: optional the unique id to use
  :return: the newly created dataset or None if an error occurred
  """
  for p in _providers():
    r = p.upload(desc, files, id)
    if r:
      return r
  return None


def update(dataset, desc, files=[]):
  """
  updates the given dataset
  :param dataset: a dataset or a dataset id
  :param desc: the dict description information
  :param files: a list of FileStorage
  :return:
  """
  old = get(dataset) if isinstance(dataset, str) else dataset
  if old is None:
    return add(desc, files)
  r = old.update(desc, files)
  return r


def remove(dataset):
  """
  removes the given dataset
  :param dataset: a dataset or a dataset id
  :return: boolean whether the operation was successful
  """
  old = get(dataset) if isinstance(dataset, str) else dataset
  if old is None:
    return False
  for p in _providers():
    if p.remove(old):
      return True
  return False


def list_idtypes():
  tmp = dict()
  for d in list_datasets():
    for idtype in d.to_idtype_descriptions():
      tmp[idtype['id']] = idtype

  # also include the known elements from the mapping graph
  mapping = get_mappingmanager()
  for idtype_id in mapping.known_idtypes():
    tmp[idtype_id] = to_idtype_description(idtype_id)
  return list(tmp.values())


def get_idmanager():
  return phovea_server.plugin.lookup('idmanager')


def get_mappingmanager():
  return phovea_server.plugin.lookup('mappingmanager')
